from __future__ import annotations

from pathlib import Path
from typing import Optional, Union

import pandas as pd

from seeq.base import util
from seeq.spy._errors import *
from seeq.spy._status import Status
from seeq.spy.workbooks.job import _pull, _push


@Status.top_level_spy_function(no_session=True)
def redo(
    job_folder: Union[str, Path],
    workbooks_df: Union[pd.DataFrame, str, list] = None,
    *,
    action: Optional[str] = None,
    hard: bool = False,
    quiet: Optional[bool] = None,
    status: Optional[Status] = None
):
    """
    Marks a set of workbooks to be redone in the specified job folder.

    Parameters
    ----------
    job_folder : {str}
        A full or partial path to the job folder.

    workbooks_df : {pd.DataFrame, str, list}
        A DataFrame containing an 'ID' column that can be used to identify the
        workbooks to affect. These IDs are based on the source system (not the
        destination system). Alternatively, you can supply a workbook ID
        directly as a str or list of strs. If omitted, all workbooks in the
        job folder are affected.

    action : str
        If supplied, limits the redo to the specified actions. You can specify
        'pull' or 'push'. If not supplied, both pull and push are affected.
        Note that 'pull' automatically includes 'push'.

    hard : bool, default False
        If true, forces re-pushing of all inventory items. Otherwise, only
        causes re-push of workbook/worksheet/workstep definitions and Topic
        content.

    quiet : bool
        If True, suppresses progress output. Note that when status is
        provided, the quiet setting of the Status object that is passed
        in takes precedence.

    status : spy.Status, optional
        If specified, the supplied Status object will be updated as the command
        progresses. It gets filled in with the same information you would see
        in Jupyter in the blue/green/red table below your code while the
        command is executed.
    """
    if not util.safe_exists(job_folder):
        raise SPyValueError(f'Job folder "{job_folder}" does not exist.')

    if workbooks_df is None:
        job_workbooks_folder = _pull.get_workbooks_folder(job_folder)
        workbooks_df = pd.DataFrame([
            {'ID': workbook_id} for _, workbook_id in _pull.walk_workbook_folders(job_workbooks_folder)
        ])

    if isinstance(workbooks_df, str):
        workbooks_df = pd.DataFrame({'ID': [workbooks_df]})
    elif isinstance(workbooks_df, list):
        workbooks_df = pd.DataFrame({'ID': workbooks_df})

    if 'ID' not in workbooks_df:
        raise SPyValueError('workbooks_df must contain an ID column.')

    if action is not None:
        if action not in ['pull', 'push']:
            raise SPyValueError('action must be "pull" or "push".')
    else:
        action = 'pull'

    status_columns = [c for c in ['ID', 'Name', 'Type', 'Workbook Type'] if c in workbooks_df]

    status.df = workbooks_df[status_columns].copy()

    if action == 'pull':
        _pull.redo(job_folder, status)

    if action == 'push':
        _push.redo(job_folder, hard, status)

    status.update(
        'Successfully marked specified items to be redone. Execute spy.job.workbooks.pull() and/or '
        'spy.job.workbooks.push() again.', Status.SUCCESS)



from __future__ import annotations

import os
from pathlib import Path
from typing import Optional, Union

import pandas as pd

from seeq import spy
from seeq.base import util
from seeq.spy import _common
from seeq.spy._errors import *
from seeq.spy._session import Session
from seeq.spy._status import Status


@Status.top_level_spy_function()
def push(
    job_folder: Union[str, Path],
    *,
    resume: bool = True,
    replace: Optional[dict] = None,
    datasource: Optional[str] = None,
    errors: Optional[str] = None,
    quiet: Optional[bool] = None,
    status: Optional[Status] = None,
    session: Optional[Session] = None
) -> pd.DataFrame:
    """
    Pulls all the data that is used by the workbooks according to the Data
    Usages sections of the data_usage.json file in the job folder.

    Parameters
    ----------
    job_folder : {str}
        A full or partial path to the job folder created by
        spy.workbooks.job.pull() and populated with data by
        spy.workbooks.job.pull_data().

    resume : bool, default True
        True if the pull should resume from where it left off, False if it
        should pull everything again.

    replace : dict, default None
        A dict with the keys 'Start' and 'End'. If provided, any existing samples
        or capsules with the start date in the provided time period will be
        replaced. The start of the time period is inclusive and the end of the
        time period is exclusive. If replace is provided but data is not
        specified, all samples/capsules within the provided time period will be
        removed.

    datasource : str, optional, default 'Seeq Data Lab'
        The name of the datasource within which to contain all the pushed items.
        Items inherit access control permissions from their datasource unless it
        has been overridden at a lower level. If you specify a datasource using
        this argument, you can later manage access control (using spy.acl functions)
        at the datasource level for all the items you have pushed.

        If you instead want access control for your items to be inherited from the
        workbook they are scoped to, specify `spy.INHERIT_FROM_WORKBOOK`.

    errors : {'raise', 'catalog'}, default 'raise'
        If 'raise', any errors encountered will cause an exception. If
        'catalog', errors will be added to a 'Result' column in the status.df
        DataFrame (errors='catalog' must be combined with
        status=<Status object>).

    quiet : bool
        If True, suppresses progress output. Note that when status is
        provided, the quiet setting of the Status object that is passed
        in takes precedence.

    status : spy.Status, optional
        If specified, the supplied Status object will be updated as the command
        progresses. It gets filled in with the same information you would see
        in Jupyter in the blue/green/red table below your code while the
        command is executed. The table itself is accessible as a DataFrame via
        the status.df property.

    session : spy.Session, optional
        If supplied, the Session object (and its Options) will be used to
        store the login session state. This is useful to log in to different
        Seeq servers at the same time or with different credentials.

    """
    data_results = load_data_results(job_folder, 'push')
    item_map = spy.workbooks.job._push.load_item_map(job_folder)
    all_usages = spy.workbooks.job.data._pull.load_data_usage(job_folder)

    data_results.loc[data_results['Result'] == 'Success', 'Result'] = 'Success: Already pushed'

    to_push = dict()
    all_ids = set()
    for data_usage in all_usages.values():
        item_dict = data_usage['Definition']
        all_ids.add(item_dict['ID'])
        if (resume and item_dict['ID'] in data_results.index
                and data_results.at[item_dict['ID'], 'Result'].startswith('Success')):
            continue

        if item_dict['ID'] not in item_map:
            raise SPyRuntimeError(f'Item {item_dict["ID"]} not found in item map, you may need to run '
                                  f'spy.workbooks.job.push() first or do the whole job over again.')

        dummy_row = item_map.dummy_items[item_map.dummy_items['ID'] == item_map[item_dict['ID']]]

        if len(dummy_row) == 0:
            continue

        dummy_dict = dummy_row.iloc[0].to_dict()

        # Parents/paths will not be pushed as part of this operation, they whould have already
        # been pushed (and relationships established) by spy.workbooks.job.push()
        dummy_dict.pop('Path', None)
        dummy_dict.pop('Asset', None)
        dummy_dict.pop('Parent ID', None)
        dummy_dict.pop('Parent Data ID', None)

        if not pd.isna(dummy_dict.get('Original ID', None)):
            to_push[dummy_dict['Original ID']] = dummy_dict

    metadata = pd.DataFrame.from_dict(to_push, orient='index')

    if 'ID' in metadata.columns:
        # Remove the ID column so that metadata is pushed via batch endpoints. (cleanse_data_ids=False is
        # used below to eliminate any tampering with Data IDs.)
        metadata.drop(columns=['ID'], inplace=True)

    def _load_df(_index):
        _pickle_filename = spy.workbooks.job.data._pull.get_df_filename(job_folder, _index)
        if not util.safe_exists(_pickle_filename):
            return None

        return pd.read_pickle(_pickle_filename)

    def _on_update(_index):
        data_results.update(status.df, overwrite=True)
        save_data_results(job_folder, data_results, 'push')

    status.on_update = _on_update
    spy.push(_load_df, metadata=metadata, workbook=spy.workbooks.job._push.get_dummy_workbook_name(job_folder),
             replace=replace, cleanse_data_ids=False, datasource=datasource, status=status, session=session)

    data_results = data_results[data_results.index.isin(all_ids)]
    data_results.update(status.df, overwrite=True)

    additional_results = status.df[~status.df.index.isin(data_results.index)]
    if len(data_results) > 0 and len(additional_results) > 0:
        data_results = pd.concat([data_results, additional_results])
    elif len(data_results) == 0:
        data_results = additional_results
    save_data_results(job_folder, data_results, 'push')

    return data_results


def get_data_results_filename(job_folder: str, prefix: str):
    return os.path.join(job_folder, f'{prefix}_data_results.pickle')


def load_data_results(job_folder: str, prefix: str) -> pd.DataFrame:
    data_results_filename = get_data_results_filename(job_folder, prefix)
    data_results: pd.DataFrame = pd.DataFrame(
        columns=['Result', 'ID', 'Type', 'Path', 'Asset', 'Name', 'Time', 'Count', 'Pages'])
    if util.safe_exists(data_results_filename):
        try:
            data_results = pd.read_pickle(data_results_filename)
        except Exception as e:
            raise SPyRuntimeError(
                f'Error loading "{data_results_filename}", please delete it manually to continue:\n{e}')

    return data_results


def save_data_results(job_folder: str, data_results: pd.DataFrame, prefix: str):
    data_results.to_pickle(get_data_results_filename(job_folder, prefix), protocol=_common.DEFAULT_PICKLE_PROTOCOL)


def redo(job_folder: str, status: Status):
    data_results = load_data_results(job_folder, 'push')
    data_results_ids = data_results['ID'].to_list()
    item_ids: pd.Series = status.df['ID']
    for index, item_id in item_ids.items():
        if item_id in data_results_ids:
            data_results.loc[data_results['ID'] == item_id, 'Result'] = 'Pushing'
            status.df.at[index, 'Result'] = 'Data push will be redone'
        else:
            status.df.at[index, 'Result'] = 'Not found'

    save_data_results(job_folder, data_results, 'push')



import copy
import csv
import hashlib
import json
import os
from dataclasses import dataclass, field
from typing import Any, List, Optional, Dict, Set

import pandas as pd

from seeq.base import util
from seeq.spy import _common
from seeq.spy._errors import *
from seeq.spy._session import Session
from seeq.spy._status import Status


class DatasourceMapList:
    _maps: List[dict]

    def __init__(self, maps: List[dict] = None):
        self._maps = maps if maps is not None else list()

    def __getitem__(self, key) -> dict:
        return self._maps.__getitem__(key)

    def __setitem__(self, key, val: dict):
        return self._maps.__setitem__(key, val)

    def __len__(self):
        return self._maps.__len__()

    def _find(self, datasource_class: str, datasource_id: str) -> int:
        for i in range(len(self._maps)):
            datasource_map = self._maps[i]
            if (datasource_map['Datasource Class'] == datasource_class and
                    datasource_map['Datasource ID'] == datasource_id):
                return i

        return -1

    def __contains__(self, datasource_map: dict) -> bool:
        return self._find(datasource_map['Datasource Class'], datasource_map['Datasource ID']) >= 0

    def get(self, datasource_class: str, datasource_id: str) -> dict:
        index = self._find(datasource_class, datasource_id)
        if index == -1:
            raise SPyValueError(f'Datasource map for Datasource Class "{datasource_class}" and Datasource ID '
                                f'"{datasource_id}" not found')

        return self._maps[index]

    def append(self, datasource_map: dict, overwrite=False):
        if not isinstance(datasource_map, dict):
            raise SPyTypeError('append() requires a dict argument')

        for key in ['Datasource Class', 'Datasource ID']:
            if key not in datasource_map:
                raise SPyValueError(f'append() datasource_map requires a {key}')

        index = self._find(datasource_map['Datasource Class'], datasource_map['Datasource ID'])
        if index >= 0:
            if overwrite:
                self._maps[index] = datasource_map
        else:
            self._maps.append(datasource_map)

    def extend(self, datasource_maps: List[dict], overwrite=False):
        if not isinstance(datasource_maps, (list, DatasourceMapList)):
            raise SPyTypeError('append() requires a list or DatasourceMapList argument')

        for datasource_map in datasource_maps:
            self.append(datasource_map, overwrite=overwrite)

    def copy(self):
        return DatasourceMapList(copy.deepcopy(self._maps))


@dataclass
class WorkbookPushContext:
    access_control: Optional[str]
    datasource: Optional[str]
    dummy_items_workbook_context: Optional[Any]
    include_annotations: Optional[bool]
    override_max_interp: Optional[bool]
    owner: Optional[str]
    reconcile_inventory_by: str
    global_inventory: str
    session: Session
    specific_worksheet_ids: Optional[List[str]]
    pushed_inventory: Optional[Dict[str, pd.DataFrame]]
    status: Status
    dry_run: bool
    datasource_maps: Optional[DatasourceMapList] = None
    failed_mappings: Set[str] = field(default_factory=set)

    def add_server_scoped_item_level_map_files(self, datasource_map_list: DatasourceMapList):
        # These "server-scoped item-level map files" serve as a sort of cache, one that
        # skips the expensive search on the server if it has been encountered before. It's
        # especially useful for datasource map folders that get re-used across multiple
        # pushes. If the datasource map definition changes at all, it would change the
        # hash in the filename and effectively "invalidate" the cache.
        for datasource_map in datasource_map_list:
            self._set_server_scoped_item_level_map_file_name(datasource_map)
            server_scoped_filename = datasource_map[_common.DATASOURCE_MAP_SERVER_SCOPED_ITEM_LEVEL_MAP_FILE]
            item_level_map_files = datasource_map.setdefault(_common.DATASOURCE_MAP_ITEM_LEVEL_MAP_FILES, list())
            item_level_map_files_set = {f.lower() for f in item_level_map_files}
            if server_scoped_filename.lower() in item_level_map_files_set:
                continue

            if util.safe_exists(server_scoped_filename):
                self.status.log(f'Datasource map "{os.path.basename(datasource_map["File"])}" has server-scoped '
                                f'item-level datasource map file: {server_scoped_filename}')
                item_level_map_files.append(server_scoped_filename)

    def _set_server_scoped_item_level_map_file_name(self, datasource_map):
        json_to_hash = copy.deepcopy(datasource_map)
        filename = json_to_hash.pop('File')
        json_to_hash.pop('DataFrames', None)
        json_to_hash.pop(_common.DATASOURCE_MAP_SERVER_SCOPED_ITEM_LEVEL_MAP_FILE, None)
        json_str = json.dumps(json_to_hash, sort_keys=True)
        hostname = self.session.public_url.replace('https://', '').replace('http://', '').replace(':', '_').lower()
        hash_object = hashlib.md5(json_str.encode())
        hash_hex = hash_object.hexdigest()[:8]
        server_scoped_filename = f'Server_Scoped_Item_Level_Datasource_Map_{hostname}_{hash_hex}.csv'
        server_scoped_fullpath = os.path.join(os.path.dirname(filename), server_scoped_filename)
        datasource_map[_common.DATASOURCE_MAP_SERVER_SCOPED_ITEM_LEVEL_MAP_FILE] = server_scoped_fullpath

    def append_to_server_scoped_item_level_map_file(self, datasource_map, source_item, destination_item):
        if _common.DATASOURCE_MAP_SERVER_SCOPED_ITEM_LEVEL_MAP_FILE not in datasource_map:
            return

        server_scoped_filename = datasource_map[_common.DATASOURCE_MAP_SERVER_SCOPED_ITEM_LEVEL_MAP_FILE]

        file_exists = util.safe_isfile(server_scoped_filename)
        with util.safe_open(server_scoped_filename, mode='a', newline='', encoding='utf-8') as csvfile:
            fieldnames = ['Old Type', 'Old ID', 'Old Datasource Class', 'Old Datasource Name', 'Old Datasource ID',
                          'Old Path', 'Old Asset', 'Old Name',
                          'New Type', 'New ID', 'New Datasource Class', 'New Datasource Name', 'New Datasource ID',
                          'New Path', 'New Asset', 'New Name']
            writer = csv.DictWriter(csvfile, fieldnames=fieldnames)

            if not file_exists:
                writer.writeheader()

            destination_datasource_name = source_item.get('Datasource Name')
            if destination_datasource_name is None and destination_item.datasource is not None:
                destination_datasource_name = destination_item.datasource.name

            writer.writerow({
                'Old Type': source_item.type,
                'Old ID': source_item.id,
                'Old Datasource Class': source_item.get('Datasource Class'),
                'Old Datasource Name': source_item.get('Datasource Name'),
                'Old Datasource ID': source_item.get('Datasource ID'),
                'Old Path': source_item.get('Path'),
                'Old Asset': source_item.get('Asset'),
                'Old Name': source_item.name,
                'New Type': destination_item.type,
                'New ID': destination_item.id,
                'New Datasource Class': destination_item.get('Datasource Class'),
                'New Datasource Name': destination_datasource_name,
                'New Datasource ID': destination_item.get('Datasource ID'),
                'New Path': destination_item.get('Path'),
                'New Asset': destination_item.get('Asset'),
                'New Name': destination_item.name
            })

        self.status.log(f'Appended item-level mapping to {os.path.basename(server_scoped_filename)}: '
                        f'{source_item.id} -> {destination_item.id}')



class ContextSwitchable:
    def __init__(self, parent):
        self._parent = parent
        self._context = self._determine_context()

    def _determine_context(self):
        parent_class_names = {cls.__name__ for cls in self._parent.__class__.__mro__}
        if "AnalysisWorksheet" in parent_class_names:
            return "worksheet"
        elif "AnalysisWorkstep" in parent_class_names:
            return "workstep"
        else:
            raise ValueError(f"Unsupported parent class: {self._parent.__class__.__name__}")

    @property
    def _getter_workstep(self):
        if self._context == "worksheet":
            return self._parent.current_workstep()
        elif self._context == "workstep":
            return self._parent
        return None

    @property
    def _setter_workstep(self):
        if self._context == "worksheet":
            return self._parent.branch_current_workstep()
        elif self._context == "workstep":
            return self._parent
        return None



import csv
import json
import logging
import re
import uuid
from typing import List

from seeq.base.seeq_names import SeeqNames
from seeq import sdk, spy
from seeq.sdk.rest import ApiException
from seeq.sdk import *
from seeq.base import util


from source import helper, shared, tables
from source.tables import ColumnSpec

GUID_REGEX = r"[0-9A-F]{8}-[0-9A-F]{4}-[0-9A-F]{4}-[0-9A-F]{4}-[0-9A-F]{12}"


def items(**kwargs):
    """
    List datasource items in use within workbooks.
    :param csv: CSV file to write items to.
    :param workbook_id: Id of the workbook whose items will be returned.
    :param workbook_csv: CSV file with workbook ids.
    :param ignore_errors: Ignore errors.
    :param include_calculations: Include formulas in the items returned.
    """
    ignore_errors = kwargs.get("ignore_errors") or False
    include_calculations = kwargs.get("include_calculations") or False

    get_workbook_items(
        kwargs.get("csv"),
        kwargs.get("workbook_id"),
        kwargs.get("workbook_csv"),
        ignore_errors,
        True,
        include_calculations,
        False,
        False,
    )
    return None


def get_workbook_items(
    local_csv,
    workbook_id,
    workbook_csv,
    ignore_errors,
    include_dependents,
    include_calculations,
    only_include_supports_caching,
    only_include_leaf_calculations,
):
    """
    This method returns a list of items in the given workbook(s)/topic(s).
    :param local_csv: CSV file to write items to.
    :param workbook_id: Id of the workbook whose items will be returned.
    :param workbook_csv: CSV file with workbook ids.
    :param ignore_errors: Ignore errors.
    :param include_dependents: Include indirectly referenced items.
    :param include_calculations: Include formulas in the items returned.
    :param only_include_supports_caching: Only include items that can be cached.
    :param only_include_leaf_calculations: Only include formulas that don't depend on other items.
           If this flag is used without the include_calculations flag, it has no effect.
    :return: List of items
    """
    client = spy.client
    items_api = sdk.ItemsApi(client)
    datasources_api = sdk.DatasourcesApi(client)
    workbooks_api = sdk.WorkbooksApi(client)

    datasources = datasources_api.get_datasources(
        limit=helper.get_default_page_size(), include_archived=True
    )  # type: DatasourceOutputListV1

    column_specs = [
        ColumnSpec("count", "#", ">", 4),
        ColumnSpec("workbook_id", "Workbook ID", console=False),
        ColumnSpec("workbook_name", "Workbook Name", "<", 36),
        ColumnSpec("workbook_owner", "Workbook Owner", "<", 25),
        ColumnSpec("workbook_type", "Workbook Type", console=False),
        ColumnSpec("updated_at", "Workbook Updated", console=False),
        ColumnSpec("name", "Item Name", "<", 40),
        ColumnSpec("type", "Item Type", console=False),
        ColumnSpec("used_in", "Used In", console=False),
        ColumnSpec("item_id", "Item ID", console=False),
        ColumnSpec("datasource", "Datasource", "<", 30),
        ColumnSpec("id_datasource", "ID (Datasource)", console=False),
        ColumnSpec("data_id", "Data ID", console=False),
        ColumnSpec("max_interp_or_duration", "Max Interp/Duration", console=False),
        ColumnSpec("formula", "Formula", console=False),
        ColumnSpec("formula_inputs", "Formula Inputs", console=False),
    ]

    table = tables.start_table(column_specs, local_csv)

    if not workbook_id and not workbook_csv:
        raise RuntimeError("You must specify either --workbook-id or --workbook-csv")

    items = []
    workbook_ids = []
    if workbook_id:
        workbook_ids = [workbook_id]
    elif workbook_csv:
        header, rows = tables.read_table(workbook_csv)

        id_column = tables.find_column_index(header, "ID")

        if id_column is None:
            raise RuntimeError('"ID" column not found in workbook CSV file')

        workbook_ids = [row[id_column] for row in rows]

    try:
        for workbook_id in workbook_ids:
            try:
                workbook = workbooks_api.get_workbook(
                    id=workbook_id
                )  # type: WorkbookOutputV1
                worksheets = workbooks_api.get_worksheets(
                    workbook_id=workbook.id
                )  # type: WorksheetOutputListV1
            except ApiException as e:
                if e.status == 403:
                    item = shared.Bunch(
                        workbook_id=workbook_id,
                        workbook_name="Access denied",
                        workbook_owner="Access denied",
                        workbook_type="Access denied",
                        updated_at="",
                        item_id="Access denied",
                        type="",
                        name="",
                        datasource="",
                        ids_datasource="",
                        data_id="",
                        used_in="",
                        max_interp_or_duration="",
                        formula="",
                        formula_inputs="",
                    )

                    tables.write_table_row(table, item)

                    if not ignore_errors:
                        raise_admin_credentials_error()

                elif not ignore_errors:
                    raise

                continue

            item_ids = dict()
            for worksheet in worksheets.worksheets:  # type: WorksheetOutputV1

                try:
                    if include_dependents:
                        item_ids.update(
                            _scrape_item_ids_from_calculated_items(
                                client, workbook_id, include_calculations
                            )
                        )
                    item_ids.update(
                        _scrape_item_ids_from_workbook(
                            client, workbook, worksheet, ignore_errors
                        )
                    )
                except Exception:
                    if not ignore_errors:
                        raise
                    else:
                        continue

            item_ids_copy = dict(item_ids)
            for item_id, _ in item_ids_copy.items():
                if include_dependents:
                    item_ids.update(
                        _scrape_item_ids_from_dependencies(client, item_id, item_ids)
                    )
            for item_id, used_in in item_ids.items():
                try:
                    item = items_api.get_item_and_all_properties(
                        id=item_id
                    )  # type: ItemOutputV1
                    item_type = item.type if item else ""
                except ApiException:
                    continue
                cachable = [
                    "StoredSignal",
                    "CalculatedSignal",
                    "StoredCondition",
                    "CalculatedCondition",
                    "CalculatedScalar",
                ]
                if item_type in cachable or not only_include_supports_caching:
                    datasource_id = _get_property(item, "Datasource ID")
                    datasource_class = _get_property(item, "Datasource Class")
                    data_id = _get_property(item, "Data ID")

                    if not include_calculations:
                        if (
                            not datasource_id
                            or not data_id
                            or datasource_id == "default"
                        ):
                            continue

                    is_generated = _get_property(item, "Is Generated")
                    if is_generated:
                        # We exclude generated items, like scalars on metrics, because they are just noise for the user
                        continue

                    datasource = _get_datasource(
                        datasources, datasource_id, datasource_class
                    )
                    datasource_name = datasource.name if datasource else ""
                    id_datasource = datasource.id if datasource else ""
                    item_name = item.name if item else ""

                    max_interp_or_duration = None
                    if "Signal" in item_type:
                        max_interp_or_duration = _get_property(
                            item, "Override Maximum Interpolation"
                        )
                        if not max_interp_or_duration:
                            max_interp_or_duration = _get_property(
                                item, "Source Maximum Interpolation"
                            )
                    elif "Condition" in item_type:
                        max_interp_or_duration = _get_property(item, "Maximum Duration")
                        if not max_interp_or_duration:
                            max_interp_or_duration = _get_property(
                                item, "maximumDuration"
                            )

                    formula = None
                    formula_inputs = None
                    if item_type.startswith("Calculated"):
                        formula = _get_property(item, "Formula")

                        inputs = []
                        calculated_item = _get_calculated_item(client, item)
                        if calculated_item is None:
                            continue
                        for (
                            parameter
                        ) in (
                            calculated_item.parameters
                        ):  # type: FormulaParameterOutputV1
                            reference = (
                                parameter.item.id
                                if parameter.item
                                else parameter.formula
                            )
                            inputs.append("$%s=%s" % (parameter.name, reference))
                        formula_inputs = "\n".join(inputs)

                    item = shared.Bunch(
                        workbook_id=workbook_id,
                        workbook_name=workbook.name,
                        workbook_owner=workbook.owner.name,
                        workbook_type=workbook.type,
                        updated_at=workbook.updated_at,
                        item_id=item_id,
                        type=item_type,
                        name=item_name,
                        datasource=datasource_name,
                        id_datasource=id_datasource,
                        data_id=data_id,
                        used_in=used_in,
                        max_interp_or_duration=max_interp_or_duration,
                        formula=formula,
                        formula_inputs=formula_inputs,
                    )

                    is_calculation = item.formula_inputs is not None
                    is_leaf_calculation = (
                        is_calculation and "$" not in item.formula_inputs
                    )
                    if (
                        (not include_calculations)
                        or (not is_calculation)
                        or (not only_include_leaf_calculations)
                        or is_leaf_calculation
                    ):
                        items.append(item)
                        tables.write_table_row(table, item)
    finally:
        if table:
            tables.finish_table(table)

    return items


def _scrape_item_ids_from_dependencies(client, item_id, item_ids):
    items_api = sdk.ItemsApi(client)

    dependencies = None
    try:
        dependencies = items_api.get_formula_dependencies(
            id=item_id
        )  # type: ItemDependencyOutputV1
    except ApiException as e:
        if e.status == 404:
            # For some reason, the item_id is unknown. So just skip it.
            print("Ignoring: %s" % e)
        else:
            print(f"API error {e}")

    if dependencies and dependencies.dependencies:
        for dependency in dependencies.dependencies:  # type: ItemParameterOfOutputV1
            if dependency.id not in item_ids:
                # We specify "Global Calc" here because the only way we'll make it to this line is if this item is a
                # dependency of calculation that wasn't already found in the calculations scoped to the workbook
                item_ids[dependency.id] = "Global Calc"

    return item_ids


def _scrape_item_ids_from_calculated_items(
    client, workbook_id, include_calculated_item
):
    items_api = sdk.ItemsApi(client)

    item_ids = dict()

    tables.iterate_over_output(
        lambda offset: items_api.search_items(
            filters=["", SeeqNames.API.Flags.exclude_globally_scoped],
            scope=[workbook_id],
            offset=offset,
            limit=helper.get_default_page_size(),
            types=[
                "CalculatedScalar",
                "CalculatedSignal",
                "CalculatedCondition",
                "Chart",
                "ThresholdMetric",
            ],
        ),
        "items",
        lambda item: _scrape_item_ids_from_calculated_item(
            client, item, item_ids, include_calculated_item
        ),
    )

    return item_ids


def _get_calculated_item(client, item):
    scalars_api = sdk.ScalarsApi(client)
    signals_api = sdk.SignalsApi(client)
    conditions_api = sdk.ConditionsApi(client)
    formulas_api = sdk.FormulasApi(client)
    metrics_api = sdk.MetricsApi(client)

    try:
        if item.type == "CalculatedScalar":
            item = scalars_api.get_scalar(id=item.id)  # type: CalculatedItemOutputV1
        elif item.type == "CalculatedSignal":
            item = signals_api.get_signal(id=item.id)  # type: SignalOutputV1
        elif item.type == "CalculatedCondition":
            item = conditions_api.get_condition(id=item.id)  # type: ConditionOutputV1
        elif item.type == "Chart":
            item = formulas_api.get_function(id=item.id)  # type: CalculatedItemOutputV1
        elif item.type == "ThresholdMetric":
            item = metrics_api.get_metric(id=item.id)  # type: ThresholdMetricOutputV1
    except Exception as e:
        return None
        print(f"Error getting calculated item: {e}")

    return item


def _scrape_item_ids_from_calculated_item(
    client, item, item_ids, include_calculated_item
):
    item = _get_calculated_item(client, item)
    if item is None:
        return

    if not item.scoped_to:
        return

    if include_calculated_item:
        item_ids[item.id] = "Workbook Calc"

    if item.type == "ThresholdMetric":
        metric = item  # type: ThresholdMetricOutputV1
        if metric.measured_item:
            item_ids[metric.measured_item.id] = "Metric Measured Item"

        if metric.bounding_condition:
            item_ids[metric.bounding_condition.id] = "Metric Bounding Condition"

        if metric.thresholds:
            for threshold in metric.thresholds:  # type: ThresholdOutputV1
                if threshold.is_generated or threshold.value:
                    continue

                item_ids[threshold.item.id] = "Metric Threshold Item"

    else:
        for parameter in item.parameters:  # type: FormulaParameterOutputV1
            if parameter.item:
                item_ids[parameter.item.id] = "Calculation"


def _scrape_item_ids_from_workstep(client, workbook_id, worksheet, workstep_id):
    workbooks_api = sdk.WorkbooksApi(client)

    item_ids = dict()

    try:
        workstep = workbooks_api.get_workstep(
            workbook_id=workbook_id, worksheet_id=worksheet.id, workstep_id=workstep_id
        )
    except ApiException as e:
        if e.status == 404:
            # Apparently it's possible for a workstep to be referenced and yet not exist
            return item_ids

        raise

    matches = re.finditer(GUID_REGEX, workstep.data, re.IGNORECASE)

    for match in matches:
        item_ids[match.group(0).upper()] = "Details (%s)" % worksheet.name

    return item_ids


def _scrape_item_ids_from_workbook(client, workbook, worksheet, ignore_errors=False):
    if _is_organizer_topic(workbook):
        return _scrape_item_ids_from_topic(client, worksheet, ignore_errors)
    else:
        return _scrape_item_ids_from_analysis(client, workbook, worksheet)


def _scrape_item_ids_from_analysis(client, workbook, worksheet):
    annotations_api = sdk.AnnotationsApi(client)

    item_ids = dict()

    if not worksheet.workstep:
        # It's possible for a workstep to have never been posted to a worksheet
        return item_ids

    workstep_id = worksheet.workstep.split("/")[-1]
    item_ids.update(
        _scrape_item_ids_from_workstep(client, workbook.id, worksheet, workstep_id)
    )

    annotations = annotations_api.get_annotations(
        annotates=[worksheet.id]
    )  # type: AnnotationListOutputV1

    for annotation in annotations.items:  # type: AnnotationOutputV1
        annotation = annotations_api.get_annotation(id=annotation.id)

        if not annotation.document:
            continue

        matches = re.finditer(
            r"item&#61;(" + GUID_REGEX + r")", annotation.document, re.IGNORECASE
        )
        for match in matches:
            item_ids[match.group(1).upper()] = "Journal (%s)" % worksheet.name

        matches = re.finditer(
            r"workstep&#61;(" + GUID_REGEX + r")", annotation.document, re.IGNORECASE
        )
        for match in matches:
            item_ids.update(
                _scrape_item_ids_from_workstep(
                    client, workbook.id, worksheet, match.group(1).upper()
                )
            )

    return item_ids


def _scrape_item_ids_from_topic(client, worksheet, ignore_errors=False):
    annotations_api = skd.AnnotationsApi(client)
    workbooks_api = sdk.WorkbooksApi(client)
    content_api = sdk.ContentApi(client)

    item_ids = dict()

    annotations = annotations_api.get_annotations(
        annotates=[worksheet.id]
    )  # type: AnnotationListOutputV1

    for annotation in annotations.items:  # type: AnnotationOutputV1
        annotation = annotations_api.get_annotation(id=annotation.id)
        all_contents: List[ContentOutputV1] = [
            content_api.get_content(id=content_id)
            for content_id in annotation.content_ids
        ]

        if not annotation.document:
            continue

        for content in all_contents:
            try:
                worksheet = workbooks_api.get_worksheet(
                    workbook_id=content.source_workbook,
                    worksheet_id=content.source_worksheet,
                )
                item_ids.update(
                    _scrape_item_ids_from_workstep(
                        client,
                        content.source_workbook,
                        worksheet,
                        content.source_workstep,
                    )
                )
                content_api.clear_image_cache(id=content.id)
            except ApiException as e:
                if not ignore_errors:
                    raise

                print(
                    'Could not get items from worksheet "%s", skipping. Exception:\n%s'
                    % (content.source_worksheet, e)
                )
                continue

    for key, value in item_ids.items():
        item_ids[key] = "Embedded Content"

    return item_ids


def _is_organizer_topic(workbook):
    return workbook.type == "Topic"


def _get_property(item, property_name):
    """
    :type item: ItemOutputV1
    """
    for prop in item.properties:  # type: PropertyOutputV1
        if prop.name == property_name:
            if prop.unit_of_measure and prop.unit_of_measure != "string":
                return prop.value + prop.unit_of_measure
            else:
                return prop.value


def _get_datasource(datasources, datasource_id, datasource_class):
    """
    :type datasources: DatasourceOutputListV1
    """
    for datasource in datasources.datasources:  # type: DatasourceOutputV1
        if (
            datasource.datasource_id == datasource_id
            and datasource.datasource_class == datasource_class
        ):
            return datasource


def _find_user(user_spec, client):
    users_api = sdk.UsersApi(client)
    users = users_api.get_users(limit=1000000)  # type: UserOutputListV1

    found_users = []
    for user in users.users:  # type: UserOutputV1
        if (
            user.name.lower() == user_spec.lower()
            or user.id.lower() == user_spec.lower()
            or user.username.lower() == user_spec.lower()
        ):
            found_users.append(user)

    if len(found_users) == 0:
        raise RuntimeError('User "%s" not found' % user_spec)

    if len(found_users) > 1:
        table = tables.tables.start_table(
            [
                ColumnSpec("name", "Name", "<", 40),
                ColumnSpec("id", "ID", "<", 36),
                ColumnSpec("username", "Username", "<", 40),
                ColumnSpec("datasource_name", "Directory", "<", 40),
            ]
        )

        for found_user in found_users:
            tables.tables.write_table_row(table, found_user)

        tables.tables.finish_table(table)

        raise RuntimeError(
            'Multiple users found that match "%s" (see above), specify ID field to narrow down.'
            % user_spec
        )

    return found_users[0]


def workbook_list(**kwargs):
    """
    List workbooks datasources with enclosing folder information.
    :param filter: [{ALL,PUBLIC,OWNER,SHARED}]
                        Filter to apply to the workbooks returned
    :param csv: CSV file to write items to.
    :param archived: Return only trashed/archived workbooks
    :param owner: Filter for particular owner (specify name, username or ID).
    :param ignore_errors: Ignore errors.
    """

    filter = kwargs.get("filter") or "OWNER"
    archived = kwargs.get("archived") or False
    ignore_errors = kwargs.get("ignore_errors") or False

    client = spy.client

    folders_api = sdk.FoldersApi(client)
    projects_api = sdk.ProjectsApi(client)
    workbooks_api = sdk.WorkbooksApi(client)

    user_id = None
    if kwargs.get("owner"):
        user_id = _find_user(kwargs.get("owner"), client).id

    types_to_search = [
        SeeqNames.Types.analysis,
        SeeqNames.Types.topic,
        SeeqNames.Types.folder,
    ]
    if user_id:
        folders = folders_api.get_folders(
            filter=filter,
            is_archived=archived,
            user_id=user_id,
            limit=100000,
            types=types_to_search,
        )  # type: WorkbenchItemOutputListV1
    else:
        folders = folders_api.get_folders(
            filter=filter, is_archived=archived, limit=100000, types=types_to_search
        )  # type: WorkbenchItemOutputListV1

    folder_stack = list()
    folder_stack.append(folders)

    content_index_stack = list()
    content_index_stack.append(0)

    column_specs = [
        ColumnSpec("count", "#", ">", 4),
        ColumnSpec("path", "Folder", ">", 36),
        ColumnSpec("name", "Name", "<", 36),
        ColumnSpec("type", "Type", "<", 10),
        ColumnSpec("owner", "Owner", "<", 30),
        ColumnSpec("id", "ID", "<", 36),
        ColumnSpec("created_at", "Created At", "<", 24, console=False),
        ColumnSpec("updated_at", "Updated At", "<", 24, console=False),
    ]

    table = tables.start_table(column_specs, kwargs.get("csv"))

    try:
        while True:
            folders = folder_stack[-1]

            if content_index_stack[-1] >= len(folders.content):
                folder_stack.pop()

                if len(folder_stack) == 0:
                    break

                content_index_stack.pop()
                content_index_stack[-1] += 1

                continue

            base_output = folders.content[content_index_stack[-1]]  # type: BaseOutput

            if base_output.type == "Folder":
                try:
                    folders = folders_api.get_folders(
                        filter=filter, folder_id=base_output.id, limit=100000
                    )  # type: WorkbenchItemOutputListV1

                    folder_stack.append(folders)

                    content_index_stack.append(0)
                except Exception:
                    content_index_stack[-1] += 1
                    if not ignore_errors:
                        raise
            elif base_output.type == "Project":
                try:
                    project = projects_api.get_project(
                        id=base_output.id
                    )  # type: ProjectOutputV1

                    owner = project.owner  # type: IdentityPreviewV1
                    item = shared.Bunch(
                        id=project.id,
                        name=project.name,
                        description=project.description,
                        type=project.type,
                        owner=owner.name,
                        created_at=project.created_at,
                        updated_at=project.updated_at,
                        path=_get_folder_path(folder_stack, content_index_stack),
                    )

                    tables.write_table_row(table, item)
                except Exception:
                    if not ignore_errors:
                        raise

                content_index_stack[-1] += 1
            else:  # Analysis or Topic (use Workbooks API)
                try:
                    workbook = workbooks_api.get_workbook(
                        id=base_output.id
                    )  # type: WorkbookOutputV1

                    owner = workbook.owner  # type: IdentityPreviewV1
                    item = shared.Bunch(
                        id=workbook.id,
                        name=workbook.name,
                        description=workbook.description,
                        type=workbook.type,
                        owner=owner.name,
                        created_at=workbook.created_at,
                        updated_at=workbook.updated_at,
                        path=_get_folder_path(folder_stack, content_index_stack),
                    )

                    tables.write_table_row(table, item)
                except Exception:
                    if not ignore_errors:
                        raise

                content_index_stack[-1] += 1
    finally:
        if table:
            tables.finish_table(table)


def _get_folder_path(folder_stack, content_index_stack):
    path = ""
    for i in range(0, len(folder_stack) - 1):
        if len(path) > 0:
            path = path + " "

        path += str(folder_stack[i].content[content_index_stack[i]].name) + " >>"

    return path


def clear_cache(**kwargs):
    """
    Clear cache of items in workbook.
    :param csv: CSV file to write items to.
    :param workbook_id: Seeq ID of workbook (either topic or analysis) to
                        clear item caches for
    :param workbook_csv: Iterate over workbooks in CSV file produced by "seeq
                        workbook list"
    :param ignore_errors: Ignore errors.
    :param recursive: Recursively clear item caches.
    """

    ignore_errors = kwargs.get("ignore_errors")
    recursive = kwargs.get("recursive")

    client = spy.client
    items_api = sdk.ItemsApi(client)

    if not kwargs.get("workbook_id") and not kwargs.get("workbook_csv"):
        raise RuntimeError("You must specify either workbook_id or workbook_csv")
    items = get_workbook_items(
        kwargs.get("csv"),
        kwargs.get("workbook_id"),
        kwargs.get("workbook_csv"),
        ignore_errors,
        recursive,
        True,
        True,
        recursive,
    )
    for item in items:
        if item.item_id:
            try:
                # Clear item caches asynchronously
                items_api.clear_cache(id=item.item_id, callback=lambda response: None)
            except Exception:
                if not ignore_errors:
                    raise


def _read_item_map(map_file):
    header, rows = _read_swap_map_csv(map_file)

    old_id_index = tables.find_column_index(header, "Old ID")
    new_id_index = tables.find_column_index(header, "New ID")

    if old_id_index is None:
        raise Exception('Column "Old ID" not found in datasource map file.')

    if new_id_index is None:
        raise Exception('Column "New ID" not found in datasource map file.')

    item_map = dict()
    item_map["header"] = header
    item_map["rows"] = rows
    item_map["old_id_index"] = old_id_index
    item_map["new_id_index"] = new_id_index

    # Weed out any non-GUIDs, like "NO MATCHES" and "MULTIPLE MATCHES"
    item_map["rows"] = [
        row
        for row in rows
        if _is_guid(row[item_map["old_id_index"]])
        and _is_guid(row[item_map["new_id_index"]])
    ]

    return item_map


def _is_guid(s):
    try:
        uuid.UUID(s)
        return True
    except Exception:
        return False


def swap(**kwargs):
    """
    Swap datasource items within the workbook for other items.
    :param map_csv: CSV file with item mapping, see datasource_map() function.
    :param workbook_id: Seeq ID of workbook to perform swap on.
    :param workbook_csv: Iterate over workbooks in CSV file produced by workbook_list().
    :param reverse: Swap in the reverse direction, new to old.
    :param include_globals: Include items referenced by globally-scoped calculations.
    :param include_dependencies: Include items references as parameters/dependencies.
    :param ignore_errors: Ignore errors encountered while accessing data.
    :param csv: Name of CSV file to route output to.
    :param append_csv: Append to the existing CSV file if possible.

    Original design doc:
    https://seeq.atlassian.net/wiki/spaces/SQ/pages/443646015/CRAB-13251+Ability+to+re-map+a+workbook+to+a+different+datasource

    Here is the sequence of actions this command executes:

    1 Iterate over all workbooks:
    1.1 Iterate over all worksheets:
    1.1.1 For Analysis workbooks:
    1.1.1.1 For the current workstep, scrape anything that looks like a GUID. Look it up in our map, and if it needs
            to be swapped for a new ID, replace it in the workstep.
    1.1.1.2 Look at the Journal text:
    1.1.1.2.1 Swap any direct "item" references as necessary.
    1.1.1.2.2 If there are workstep links, scrape & swap that workstep as we did in 1.1.1.1
    1.1.1.3 Search for calculated items that are scoped to the workbook. Swap any of their parameters as necessary.
            CalculatedScalar, CalculatedSignal, CalculatedCondition and Chart (histogram) all need special
            treatment.
    1.1.2 For Topic workbooks:
    1.1.2.1 Look for embedded content in the Topic document and scrape & swap the workstep as we did in 1.1.1.1.
    1.1.2.2 Remove the <img src> attributes for all embedded content found in 1.1.2.2.
    2 Optionally iterate over globally-scoped calculated items. Swap as necessary similar to 1.1.1.3.
    """
    item_map = _read_item_map(kwargs.get("map_csv"))

    kwargs["reverse"] = kwargs.get("reverse") or False
    kwargs["include_globals"] = kwargs.get("include_globals") or False
    kwargs["include_dependencies"] = kwargs.get("include_dependencies") or False
    kwargs["ignore_errors"] = kwargs.get("ignore_errors") or False

    logging.basicConfig(
        filename="workbook_swap.log",
        level=logging.INFO,
        format="%(asctime)s - %(message)s",
    )

    client = spy.client

    workbooks_api = sdk.WorkbooksApi(client)
    items_api = sdk.ItemsApi(client)

    column_specs = [
        ColumnSpec("count", "#", ">", 4),
        ColumnSpec("workbook_id", "Workbook ID", "<", 36),
        ColumnSpec("workbook_name", "Workbook Name", "<", 36),
        ColumnSpec("workbook_type", "Workbook Type", "<", 13, console=False),
        ColumnSpec("old_item_id", "Old Item ID", "<", 36, console=False),
        ColumnSpec("new_item_id", "New Item ID", "<", 36),
        ColumnSpec("used_in", "Used In", "<", 16),
    ]

    if not kwargs.get("workbook_id") and not kwargs.get("workbook_csv"):
        raise RuntimeError("You must specify either workbook_id or workbook_csv")

    workbook_ids = []
    if kwargs.get("workbook_id"):
        workbook_ids = [kwargs.get("workbook_id")]
    elif kwargs.get("workbook_csv"):
        header, rows = tables.read_table(kwargs.get("workbook_csv"))

        id_column = tables.find_column_index(header, "ID")

        if id_column is None:
            raise RuntimeError('"ID" column not found in workbook CSV file')

        workbook_ids = [row[id_column] for row in rows]

    table = tables.start_table(
        column_specs,
        csv_file_name=kwargs.get("csv"),
        append_csv=kwargs.get("append_csv"),
    )

    try:
        for workbook_id in workbook_ids:
            log("Processing workbook %s" % workbook_id)
            try:
                workbook = workbooks_api.get_workbook(
                    id=workbook_id
                )  # type: WorkbookOutputV1
                worksheets = workbooks_api.get_worksheets(
                    workbook_id=workbook.id,  # type: WorksheetOutputListV1
                    limit=helper.get_default_page_size(),
                )
            except ApiException as e:
                if e.status == 403:
                    tables.write_table_row(
                        table,
                        shared.Bunch(
                            workbook_id=workbook_id,
                            workbook_name="Access Denied",
                            workbook_type="",
                            used_in="",
                            old_item_id="",
                            new_item_id="",
                        ),
                    )

                    if not kwargs.get("ignore_errors"):
                        raise_admin_credentials_error()

                    continue

                if not kwargs.get("ignore_errors"):
                    raise

                tables.write_table_row(
                    table,
                    shared.Bunch(
                        workbook_id=workbook_id,
                        workbook_name="Error: %s" % e,
                        workbook_type="",
                        used_in="",
                        old_item_id="",
                        new_item_id="",
                    ),
                )
                continue

            for worksheet in worksheets.worksheets:  # type: WorksheetOutputV1
                log(
                    "Processing worksheet %s (%s)" % (worksheet.id, worksheet.name),
                    indentation_level=1,
                )
                # This swaps non-calculated items in the details pane of the current worksheet
                try:
                    if workbook.type == "Analysis":
                        _swap_analysis_worksheet(
                            client,
                            table,
                            item_map,
                            workbook,
                            workbook.type,
                            worksheet,
                            kwargs.get("reverse"),
                        )
                    else:
                        _swap_topic_document(
                            client, table, item_map, worksheet, kwargs.get("reverse")
                        )
                except ApiException as e:
                    if not kwargs.get("ignore_errors"):
                        raise

                    tables.write_table_row(
                        table,
                        shared.Bunch(
                            workbook_id=workbook_id,
                            workbook_name="Error: %s" % e,
                            workbook_type="",
                            used_in="",
                            old_item_id=worksheet.id,
                            new_item_id="",
                        ),
                    )
                    continue

                # This swaps global calculated items present in the details pane of the worksheet
                if kwargs.get("include_globals"):
                    # If we're supposed to include globals that are referenced by this workbook, first scrape all
                    # item IDs and feed them into _swap_calculated_item() with allow_globals=True
                    item_ids = _scrape_item_ids_from_workbook(
                        client, workbook, worksheet, kwargs.get("ignore_errors")
                    )  # type: dict
                    log("Swapping global calculated items", indentation_level=1)
                    for item_id in item_ids.keys():
                        try:
                            item = items_api.get_item_and_all_properties(
                                id=item_id
                            )  # type: ItemOutputV1
                            if item.scoped_to:
                                # Ignore non-global calculated items because they'll be processed later below
                                continue

                            _swap_calculated_item(
                                client,
                                table,
                                item_map,
                                workbook,
                                workbook.type,
                                item,
                                kwargs.get("reverse"),
                                allow_globals=True,
                                ignore_errors=kwargs.get("ignore_errors"),
                                include_dependencies=kwargs.get("include_dependencies"),
                                log_indentation_level=2,
                            )
                        except ApiException as e:
                            if not kwargs.get("ignore_errors"):
                                raise

                            tables.write_table_row(
                                table,
                                shared.Bunch(
                                    workbook_id=workbook_id,
                                    workbook_name="Error: %s" % e,
                                    workbook_type="",
                                    used_in="",
                                    old_item_id=item_id,
                                    new_item_id="",
                                ),
                            )
                            continue

            log(
                "Swapping calculated items for workbook %s" % workbook.id,
                indentation_level=1,
            )
            try:
                tables.iterate_over_output(
                    lambda offset: items_api.search_items(
                        filters=[
                            "",
                            SeeqNames.API.Flags.exclude_globally_scoped,
                            SeeqNames.API.Flags.include_unsearchable,
                        ],
                        scope=[workbook_id],
                        offset=offset,
                        limit=helper.get_default_page_size(),
                        types=[
                            "CalculatedScalar",
                            "CalculatedSignal",
                            "CalculatedCondition",
                            "Chart",
                            "ThresholdMetric",
                        ],
                    ),
                    "items",
                    lambda item: _swap_calculated_item(
                        client,
                        table,
                        item_map,
                        workbook,
                        workbook.type,
                        item,
                        kwargs.get("reverse"),
                        allow_globals=False,
                        ignore_errors=kwargs.get("ignore_errors"),
                        include_dependencies=kwargs.get("include_dependencies"),
                        log_indentation_level=2,
                    ),
                )
            except Exception as e:
                log(
                    'Could not retrieve data for workbook "%s", skipping. Exception:\n%s'
                    % (workbook_id, e),
                    indentation_level=0,
                )

    finally:
        tables.finish_table(table)


def _swap_topic_document(client, table, item_map, worksheet, reverse):
    log(
        "Swap topic document %s (%s)" % (worksheet.id, worksheet.name),
        indentation_level=2,
    )

    workbooks_api = sdk.WorkbooksApi(client)
    annotations_api = sdk.AnnotationsApi(client)
    content_api = sdk.ContentApi(client)

    annotations = annotations_api.get_annotations(
        annotates=[worksheet.id]
    )  # type: AnnotationListOutputV1

    for annotation in annotations.items:  # type: AnnotationOutputV1
        annotation = annotations_api.get_annotation(id=annotation.id)

        if not annotation.document:
            continue

        all_contents: List[ContentOutputV1] = [
            content_api.get_content(id=content_id)
            for content_id in annotation.content_ids
        ]

        for content in all_contents:
            embedded_workbook = workbooks_api.get_workbook(id=content.source_workbook)
            embedded_worksheet = workbooks_api.get_worksheet(
                workbook_id=content.source_workbook,
                worksheet_id=content.source_worksheet,
            )

            _swap_workstep(
                client,
                table,
                item_map,
                embedded_workbook,
                "Analysis",
                embedded_worksheet,
                content.source_workstep,
                reverse,
                used_in="Embedded Content",
                log_indentation_level=3,
            )

            content_api.clear_image_cache(id=content.id)


def _update_annotation_document(client, annotation, new_document):
    """
    :type annotation: AnnotationOutputV1
    """
    annotations_api = sdk.AnnotationsApi(client)

    new_annotation = AnnotationInputV1()
    new_annotation.discoverable = annotation.discoverable
    new_annotation.document = new_document
    new_annotation.name = annotation.name
    new_annotation.description = annotation.description
    new_annotation.interests = []
    for interest in annotation.interests:  # type: AnnotationInterestOutputV1
        interest_item = interest.item  # type: ItemPreviewV1
        # we've encountered a case where there were multiple interests returned with the same ID, which
        # caused Appserver to choke when updating the annotation. So filter those out.
        if any(interest_item.id == i.interest_id for i in new_annotation.interests):
            continue
        new_interest = AnnotationInterestInputV1()
        new_interest.interest_id = interest_item.id
        if interest.capsule is not None:
            new_interest.detail_id = interest.capsule.id
        new_annotation.interests.append(new_interest)
    new_annotation.type = annotation.type
    if annotation.type == "Report":
        new_annotation.report_input = OptionalReportInputV1()
    new_annotation.created_by_id = annotation.created_by.id

    annotations_api.update_annotation(id=annotation.id, body=new_annotation)


def _swap_analysis_worksheet(
    client, table, item_map, workbook, workbook_type, worksheet, reverse
):
    annotations_api = sdk.AnnotationsApi(client)

    log(
        "Swap analysis worksheet %s (%s)" % (worksheet.id, worksheet.name),
        indentation_level=2,
    )

    # Short-circuit if this worksheet does not have a workstep
    if not worksheet.workstep:
        return

    # Gets the current workstep id from a string that looks like:
    # "workbook/WORKBOOK_GUID/worksheet/WORKSHEET_GUID/workstep/WORKSTEP_GUID"
    current_workstep_id = worksheet.workstep.split("/")[-1]

    # Searches the .data property of the workstep using a regular expression and swaps GUIDs if they are present in
    # the mapping CSV (supplied as an argument)
    # Effectively performs a swap on non-calculated items in the details pane of the worksheet
    _swap_workstep(
        client,
        table,
        item_map,
        workbook,
        workbook_type,
        worksheet,
        current_workstep_id,
        reverse,
        log_indentation_level=3,
    )

    # Now it is time to swap annotations!
    annotations = annotations_api.get_annotations(
        annotates=[worksheet.id]
    )  # type: AnnotationListOutputV1

    log(
        "Swap annotations for analysis worksheet %s (%s)"
        % (worksheet.id, worksheet.name),
        indentation_level=2,
    )
    for annotation in annotations.items:  # type: AnnotationOutputV1
        annotation = annotations_api.get_annotation(
            id=annotation.id
        )  # type: AnnotationOutputV1
        log("Swap annotation %s " % annotation.id, indentation_level=3)

        if not annotation.document:
            continue

        annotation_modified = False
        new_document = annotation.document

        # Obtains an iterator over all GUIDs in the annotation
        matches = re.finditer(
            r"item&#61;(" + GUID_REGEX + r")", annotation.document, re.IGNORECASE
        )
        for match in matches:
            # If the GUID should be swapped, swap it.
            row_index = _lookup_old_id_in_map(match.group(1).upper(), item_map, reverse)

            if row_index is not None:
                old_id = item_map["rows"][row_index][
                    item_map["new_id_index"] if reverse else item_map["old_id_index"]
                ]
                new_id = item_map["rows"][row_index][
                    item_map["old_id_index"] if reverse else item_map["new_id_index"]
                ]

                new_document = new_document.replace(match.group(1), new_id)

                annotation_modified = True

                tables.write_table_row(
                    table,
                    shared.Bunch(
                        workbook_id=workbook.id,
                        workbook_name=workbook.name,
                        workbook_type=workbook_type,
                        used_in="Journal",
                        old_item_id=old_id,
                        new_item_id=new_id,
                    ),
                )

        if annotation_modified:
            log("Updating annotation document.", indentation_level=3)
            _update_annotation_document(client, annotation, new_document)
        else:
            log("... annotation not modified.", indentation_level=3)

        # If the annotation links to another workstep, perform the swap on that workstep too
        matches = re.finditer(
            r"workstep&#61;(" + GUID_REGEX + r")", annotation.document, re.IGNORECASE
        )
        for match in matches:
            workstep_id = match.group(1)
            _swap_workstep(
                client,
                table,
                item_map,
                workbook,
                workbook_type,
                worksheet,
                workstep_id,
                reverse,
                log_indentation_level=4,
            )


def _swap_workstep(
    client,
    table,
    item_map,
    workbook,
    workbook_type,
    worksheet,
    workstep_id,
    reverse,
    used_in="Details Pane",
    log_indentation_level=0,
):
    workbooks_api = sdk.WorkbooksApi(client)
    items_api = sdk.ItemsApi(client)

    # Get this workstep via API call
    try:
        workstep = workbooks_api.get_workstep(
            workbook_id=workbook.id, worksheet_id=worksheet.id, workstep_id=workstep_id
        )  # type: WorkstepOutputV1
    except ApiException as e:
        if e.status == 404:
            # Apparently it's possible for a workstep to be referenced and yet not exist
            return

        raise

    log(
        "[[ START: Swapping workstep %s" % workstep.id,
        indentation_level=log_indentation_level,
    )
    items_swapped = False  # Becomes true if at least one item is swapped
    new_data = workstep.data  # this will be modified with swapped GUIDs later

    # Iterate over all GUIDs present in the .data property of the workstep
    matches = re.finditer(GUID_REGEX, workstep.data, re.IGNORECASE)
    for match in matches:
        log(
            "* Found GUID %s in workstep data" % match.group(0).upper(),
            indentation_level=log_indentation_level + 1,
        )
        # row_index will be assigned a value if this GUID should be swapped
        row_index = _lookup_old_id_in_map(match.group(0).upper(), item_map, reverse)

        if row_index is None:
            log(
                "... no mapping found for GUID",
                indentation_level=log_indentation_level + 2,
            )
        else:
            log(
                "... mapping found for GUID. Row index in map: %d" % row_index,
                indentation_level=log_indentation_level + 2,
            )

        if row_index is not None:
            # Swap the GUIDs and store the result in new_data
            old_id = item_map["rows"][row_index][
                item_map["new_id_index"] if reverse else item_map["old_id_index"]
            ]
            new_id = item_map["rows"][row_index][
                item_map["old_id_index"] if reverse else item_map["new_id_index"]
            ]
            log(
                "Replacing %s -> %s" % (old_id, new_id),
                indentation_level=log_indentation_level + 1,
            )
            new_data = new_data.replace(match.group(0), new_id)

            items_swapped = True

            # Write the output to the console and result CSV file
            tables.write_table_row(
                table,
                shared.Bunch(
                    workbook_id=workbook.id,
                    workbook_name=workbook.name,
                    workbook_type=workbook_type,
                    used_in=used_in,
                    old_item_id=old_id,
                    new_item_id=new_id,
                ),
            )
    if items_swapped:
        # If the user is swapping assets from Area A to Area B, a situation may arise where the user already has 1 or
        # more Area B assets in the current worksheet. Swapping an Area A asset to Area B could result in a duplicate.
        # This removes these duplicates.
        new_data = _dedup_details_pane_items(new_data)

        # Upload the changes.
        property_input = PropertyInputV1()
        property_input.value = new_data
        log(
            "END ]] - Items were swapped. Setting Data property for workstep %s"
            % workstep.id,
            indentation_level=log_indentation_level,
        )
        items_api.set_property(
            id=workstep.id, property_name="Data", body=property_input
        )
    else:
        log(
            "END ]] - No items swapped for workstep %s" % workstep.id,
            indentation_level=log_indentation_level,
        )


def _dedup_details_pane_items(new_data):
    new_data_json = json.loads(new_data)
    if (
        "state" not in new_data_json
        or "stores" not in new_data_json["state"]
        or "sqTrendSeriesStore" not in new_data_json["state"]["stores"]
        or "items" not in new_data_json["state"]["stores"]["sqTrendSeriesStore"]
    ):
        return new_data

    items = new_data_json["state"]["stores"]["sqTrendSeriesStore"]["items"]
    new_items = []
    for item in items:
        if sum(i["id"] == item["id"] for i in new_items) == 0:
            new_items.append(item)

    new_data_json["state"]["stores"]["sqTrendSeriesStore"]["items"] = new_items

    return json.dumps(new_data_json)


def _swap_calculated_item(
    client,
    table,
    item_map,
    workbook,
    workbook_type,
    item,
    reverse,
    allow_globals,
    ignore_errors,
    include_dependencies,
    log_indentation_level,
):
    scalars_api = sdk.ScalarsApi(client)
    signals_api = sdk.SignalsApi(client)
    conditions_api = sdk.ConditionsApi(client)
    formulas_api = sdk.FormulasApi(client)
    metrics_api = sdk.MetricsApi(client)
    items_api = sdk.ItemsApi(client)

    log(
        "[[ START Swapping calculated item %s (%s, %s)"
        % (item.id, item.type, item.name),
        indentation_level=log_indentation_level,
    )

    items = [item]  # always include the current item
    # Add all of the dependencies to the items list
    if include_dependencies:
        item_ids = dict()
        item_ids = _scrape_item_ids_from_dependencies(client, item.id, item_ids)

        # add the corresponding items
        for item_id in item_ids.keys():
            try:
                items.append(items_api.get_item_and_all_properties(id=item_id))
            except ApiException as e:
                if not ignore_errors:
                    raise
                log(
                    'Could not retrieve data for calculated item "%s", skipping. Exception:\n%s'
                    % (item.id, e),
                    indentation_level=log_indentation_level,
                )

    # Iterate over the current item and all of its dependencies
    for item in items:
        try:
            # Determine the type of item we are considering
            if item.type == "CalculatedScalar":
                item = scalars_api.get_scalar(
                    id=item.id
                )  # type: CalculatedItemOutputV1
            elif item.type == "CalculatedSignal":
                item = signals_api.get_signal(id=item.id)  # type: SignalOutputV1
            elif item.type == "CalculatedCondition":
                item = conditions_api.get_condition(
                    id=item.id
                )  # type: ConditionOutputV1
            elif item.type == "Chart":
                item = formulas_api.get_function(
                    id=item.id
                )  # type: CalculatedItemOutputV1
            elif item.type == "ThresholdMetric":
                item = metrics_api.get_metric(
                    id=item.id
                )  # type: ThresholdMetricOutputV1
            else:
                continue

            log(
                "Swapping parameters for item %s (%s, %s)"
                % (item.id, item.type, item.name),
                indentation_level=log_indentation_level + 1,
            )
            if item.scoped_to or allow_globals:
                if item.type == "ThresholdMetric":
                    _swap_threshold_metric_parameters(
                        client,
                        table,
                        item_map,
                        workbook,
                        workbook_type,
                        item,
                        reverse,
                        log_indentation_level=log_indentation_level + 2,
                    )
                else:
                    _swap_calculated_item_parameters(
                        client,
                        table,
                        item_map,
                        workbook,
                        workbook_type,
                        item,
                        reverse,
                        log_indentation_level=log_indentation_level + 2,
                    )
        except ApiException as e:
            if not ignore_errors:
                raise

            log(
                'Could not swap calculated item "%s", skipping. Exception:\n%s'
                % (item.id, e),
                indentation_level=log_indentation_level + 1,
            )
    log(
        "END ]] - Swapping finished for calculated item %s (%s, %s)"
        % (item.id, item.type, item.name),
        indentation_level=log_indentation_level,
    )


def _swap_threshold_metric_parameters(
    client,
    table,
    item_map,
    workbook,
    workbook_type,
    old_item,
    reverse,
    log_indentation_level,
):
    """
    :type old_item:  ThresholdMetricOutputV1
    """
    metrics_api = sdk.MetricsApi(client)

    metric_modified = False

    log(
        "Swapping threshold metric parameters for %s (%s)"
        % (old_item.id, old_item.name),
        indentation_level=log_indentation_level,
    )

    new_item = ThresholdMetricInputV1()
    new_item.name = old_item.name
    if old_item.duration:
        new_item.duration = str(old_item.duration.value) + old_item.duration.uom
    if old_item.period:
        new_item.period = str(old_item.period.value) + old_item.period.uom
    new_item.scoped_to = old_item.scoped_to
    new_item.aggregation_function = old_item.aggregation_function

    if old_item.measured_item:
        row_index = _lookup_old_id_in_map(old_item.measured_item.id, item_map, reverse)
        if row_index is not None:
            new_item.measured_item = item_map["rows"][row_index][
                item_map["old_id_index"] if reverse else item_map["new_id_index"]
            ]

            metric_modified = True
            tables.write_table_row(
                table,
                shared.Bunch(
                    workbook_id=workbook.id,
                    workbook_name=workbook.name,
                    workbook_type=workbook_type,
                    used_in="Metric Measured Item",
                    old_item_id=old_item.measured_item.id,
                    new_item_id=new_item.measured_item,
                ),
            )
        else:
            new_item.measured_item = old_item.measured_item.id

    if old_item.bounding_condition:
        row_index = _lookup_old_id_in_map(
            old_item.bounding_condition.id, item_map, reverse
        )
        if row_index is not None:
            new_item.bounding_condition = item_map["rows"][row_index][
                item_map["old_id_index"] if reverse else item_map["new_id_index"]
            ]

            metric_modified = True
            tables.write_table_row(
                table,
                shared.Bunch(
                    workbook_id=workbook.id,
                    workbook_name=workbook.name,
                    workbook_type=workbook_type,
                    used_in="Metric Bounding Condition",
                    old_item_id=old_item.bounding_condition.id,
                    new_item_id=new_item.bounding_condition,
                ),
            )
        else:
            new_item.bounding_condition = old_item.bounding_condition.id

    new_item.thresholds = []  # type: str
    if old_item.thresholds:
        for threshold in old_item.thresholds:  # type: ThresholdOutputV1
            priority = threshold.priority  # type: PriorityV1
            if threshold.value:
                if isinstance(threshold.value, ScalarValueOutputV1):
                    scalar_value = threshold.value  # type: ScalarValueOutputV1
                    val = "%s%s" % (
                        scalar_value.value,
                        scalar_value.uom if scalar_value.uom else "",
                    )
                    new_item.thresholds.append("%s=%s" % (priority.level, val))
                else:
                    new_item.thresholds.append(
                        "%s=%s" % (priority.level, threshold.value)
                    )
            else:
                new_threshold_id = threshold.item.id
                row_index = _lookup_old_id_in_map(threshold.item.id, item_map, reverse)

                if row_index is not None:
                    new_threshold_id = item_map["rows"][row_index][
                        (
                            item_map["old_id_index"]
                            if reverse
                            else item_map["new_id_index"]
                        )
                    ]

                    metric_modified = True
                    tables.write_table_row(
                        table,
                        shared.Bunch(
                            workbook_id=workbook.id,
                            workbook_name=workbook.name,
                            workbook_type=workbook_type,
                            used_in="Metric Threshold",
                            old_item_id=threshold.item.id,
                            new_item_id=new_threshold_id,
                        ),
                    )

                new_item.thresholds.append("%s=%s" % (priority.level, new_threshold_id))

    if metric_modified:
        log("... metric modified", indentation_level=log_indentation_level + 1)
        metrics_api.put_threshold_metric(id=old_item.id, body=new_item)
    else:
        log("... metric not modified", indentation_level=log_indentation_level + 1)


def _swap_calculated_item_parameters(
    client,
    table,
    item_map,
    workbook,
    workbook_type,
    old_item,
    reverse,
    log_indentation_level,
):
    items_api = sdk.ItemsApi(client)
    formulas_api = sdk.FormulasApi(client)

    parameters_swapped = 0

    # For Signals, Conditions and Scalars
    formula_input = FormulaUpdateInputV1()
    formula_input.formula = old_item.formula
    formula_input.parameters = []

    # For Histograms
    function_input = FunctionInputV1()
    function_input.name = old_item.name
    function_input.formula = old_item.formula
    function_input.parameters = []

    for old_parameter in old_item.parameters:  # type: FormulaParameterOutputV1
        row_index = None
        if old_parameter.item:
            row_index = _lookup_old_id_in_map(old_parameter.item.id, item_map, reverse)

        formula_parameter_input = FormulaParameterInputV1()
        formula_parameter_input.name = old_parameter.name
        formula_parameter_input.formula = old_parameter.formula
        formula_parameter_input.unbound = old_parameter.unbound

        if row_index is not None:
            log(
                "Mapping found for %s. Row index is %d"
                % (old_parameter.item.id, row_index),
                indentation_level=log_indentation_level + 1,
            )
            old_id = item_map["rows"][row_index][
                item_map["new_id_index"] if reverse else item_map["old_id_index"]
            ]
            new_id = item_map["rows"][row_index][
                item_map["old_id_index"] if reverse else item_map["new_id_index"]
            ]
            log(
                "Swapping parameter %s -> %s" % (old_id, new_id),
                indentation_level=log_indentation_level + 1,
            )
            if old_item.type == "Chart":
                formula_parameter_input.id = new_id
                function_input.parameters.append(formula_parameter_input)
            else:
                formula_input.parameters.append(old_parameter.name + "=" + new_id)

            tables.write_table_row(
                table,
                shared.Bunch(
                    workbook_id=workbook.id,
                    workbook_name=workbook.name,
                    workbook_type=workbook_type,
                    used_in="Calculation",
                    old_item_id=old_id,
                    new_item_id=new_id,
                ),
            )

            parameters_swapped += 1

        else:
            if old_parameter.item:
                log(
                    "... mapping NOT found for %s." % old_parameter.item.id,
                    indentation_level=log_indentation_level + 1,
                )
            if old_item.type == "Chart":
                formula_parameter_input.id = (
                    old_parameter.item.id if old_parameter.item else None
                )
                function_input.parameters.append(formula_parameter_input)
            else:
                formula_input.parameters.append(
                    old_parameter.name + "=" + old_parameter.item.id
                )

    if parameters_swapped > 0:
        log(
            "%d parameters swapped. Setting the new formula" % parameters_swapped,
            indentation_level=log_indentation_level,
        )
        try:
            if old_item.type == "Chart":
                formulas_api.update_function(id=old_item.id, body=function_input)
            else:
                items_api.set_formula(id=old_item.id, body=formula_input)
        except ApiException as e:
            log(
                'Could not swap calculated item "%s", skipping. Exception:\n%s'
                % (old_item.id, e),
                indentation_level=log_indentation_level,
            )


def _lookup_old_id_in_map(old_id, item_map, reverse):
    old_id_index = item_map["new_id_index"] if reverse else item_map["old_id_index"]
    new_id_index = item_map["old_id_index"] if reverse else item_map["new_id_index"]
    for i in range(0, len(item_map["rows"])):
        if item_map["rows"][i][old_id_index] == old_id:
            # If it's not a guid, then it's not mapped or there were multiple matches
            if re.match(GUID_REGEX, item_map["rows"][i][new_id_index]):
                return i

    return None


def _read_swap_map_csv(map_filename):
    header = None
    rows = []
    with util.safe_open(map_filename, "r", encoding="utf-8-sig") as csv_file:
        csv_reader = csv.reader(csv_file)
        for row in csv_reader:
            if not header:
                header = row
            else:
                rows.append(row)

    return header, rows


def log(msg, indentation_level=0):
    if indentation_level > 0:
        msg = ".." * indentation_level + msg

    logging.info(msg)



from __future__ import annotations

from typing import Optional, List

from seeq import spy
from seeq.spy import _version, Session, Status, _datalab
from seeq.spy._errors import *


@Status.top_level_spy_function(validate_login=False)
def upgrade(
    version: Optional[str] = None,
    force_restart: bool = False,
    use_testpypi: bool = False,
    status: Optional[Status] = None,
    session: Optional[Session] = None,
    dependencies: Optional[List[str]] = None
):
    """
    Upgrades to the latest version of SPy that is compatible with this version of Seeq Server.

    An internet connection is required since this uses pip to pull the latest version from PyPI. This must be
    invoked from a Jupyter notebook or other IPython-compatible environment.

    Parameters
    ----------
    version : str, optional
        Attempts to upgrade to the provided version exactly as specified. The full SPy version must be
        provided (e.g. 221.13). For Seeq versions prior to R60, You must specify the full "seeq" module
        version (e.g. 58.0.2.184.12).

    force_restart : bool, optional
        If True, forces the kernel to shut down and restart after the upgrade. All in-memory variables and
        imports will be lost.

    use_testpypi : bool, optional
        For Seeq internal testing only.

    status : spy.Status, optional
        If specified, the supplied Status object will be updated as the command progresses. It gets filled
        in with the same information you would see in Jupyter in the blue/green/red table below your code
        while the command is executed.

    session : spy.Session, optional
        If supplied, the Session object (and its Options) will be used to store the login session state.
        This is used to access the server's current version.

    dependencies : List[str], optional
        Valid options include ["templates", "widgets", "jobs", "jupyter", "all"]
        If specified, which specific extra dependencies to install. If not specified, the installed packages
        will be scanned, and existing extras will be updated automatically. If an empty list is provided, no
        extras will be updated.

    Examples
    --------
    Upgrade to the latest version of SPy compatible with your Seeq server's major version.
    >>> spy.upgrade()

    Upgrade to version '221.13' of SPy.
    >>> spy.upgrade(version='221.13')

    """
    if session.client is None:
        raise SPyRuntimeError('Not logged in. Execute spy.login() before calling this function so that the upgrade '
                              'mechanism knows what version of Seeq Server you are interfacing with.')

    pip_command = _version.generate_pip_upgrade_command(session, version, use_testpypi, dependencies)

    try:
        import IPython
    except ImportError:
        print(f'Unable to import `IPython`. Please run `{pip_command}` in a terminal to upgrade SPy.')
        return

    ipython = IPython.get_ipython()
    if not _datalab.is_ipython() or not _datalab.is_ipython_interactive() or not ipython:
        raise SPyValueError(f'spy.upgrade() must be invoked from a Jupyter notebook or other IPython-compatible '
                            f'environment. Unable to run "{pip_command}".')
    restart_message = 'The kernel will automatically be shut down afterward.' if force_restart else \
        'Please restart the kernel once the packages have been upgraded.'
    status.update(f'Running "{pip_command}". {restart_message}', Status.RUNNING)
    ipython.run_cell(pip_command)

    if force_restart:
        if not ipython.kernel:
            raise SPyValueError(f'Unable get IPython kernel to complete restart')
        ipython.kernel.do_shutdown(True)



from __future__ import annotations

import re
from typing import List, Optional, Tuple

from seeq import spy, sdk
from seeq.spy import _common
from seeq.spy._dependencies import Dependencies
from seeq.spy._errors import *
from seeq.spy._session import Session

SDK_MODULE_VERSION_REGEX = re.compile(r'^(\d+)\.(\d+)\.(\d+).*')
SPY_MODULE_VERSION_REGEX = re.compile(r'^(?:.*\.)?(\d+)\.(\d+)$')
SEEQ_SERVER_VERSION_REGEX = re.compile(r'^R?(?:\d+\.)?(\d+)\.(\d+)\.(\d+)(-v\w+)?(-[-\w]+)?')


def get_sdk_module_version_tuple() -> Tuple[int, int, int]:
    """
    Provides a tuple of (major, minor, patch) version of Seeq SDK module (as integers).

    The major version of the Seeq SDK should match the major version of the Seeq Server.
    You can retrieve the version of the Seeq Server (once you've logged in) via
    spy.utils.get_server_version_tuple().

    Use this function instead of parsing sdk.__version__.

    Returns
    -------
    Tuple of (major, minor, patch) version of Seeq SDK module (as integers).
    """
    match = SDK_MODULE_VERSION_REGEX.match(sdk.__version__ if hasattr(sdk, '__version__') else spy.__version__)
    seeq_module_major = int(match.group(1))
    seeq_module_minor = int(match.group(2))
    seeq_module_patch = int(match.group(3))
    return seeq_module_major, seeq_module_minor, seeq_module_patch


def get_spy_module_version_tuple() -> Tuple[int, int]:
    """
    Provides a tuple of (major, minor) version of Seeq SPy module (as
    integers).

    Use this function instead of parsing spy.__version__.

    Returns
    -------
    Tuple of (major, minor) version of Seeq SPy module (as integers).
    """
    match = SPY_MODULE_VERSION_REGEX.match(spy.__version__)
    spy_module_major = int(match.group(1))
    spy_module_minor = int(match.group(2))
    return spy_module_major, spy_module_minor


def get_server_version_tuple(session: Session) -> Tuple[int, int, int]:
    """
    Provides a tuple of (major, minor, patch) version of the Seeq Server the
    supplied session is connected to (as integers). If a session is not
    supplied, the default session is used.

    The major version of the Seeq SDK should match the major version of the
    Seeq Server. You can retrieve the version of the Seeq Server (once you've
    logged in) via spy.utils.get_server_version_tuple().

    Use this function instead of parsing sdk.__version__.

    Parameters
    ----------
    session : spy.Session, optional
        If supplied, the Session object used for the connection to Seeq
        Server. If not supplied, the default session is used.

    Returns
    -------
    Tuple of (major, minor, patch) version of Seeq Server (as integers).
    """
    _common.validate_argument_types([
        (session, 'session', Session)
    ])
    session = Session.validate(session)
    if session.server_version is None:
        raise SPyRuntimeError('Not logged in. You must be logged in to a Seeq Server and have a valid session in '
                              'order to execute this function.')

    match = SEEQ_SERVER_VERSION_REGEX.match(session.server_version)
    seeq_server_major = int(match.group(1))
    seeq_server_minor = int(match.group(2))
    seeq_server_patch = int(match.group(3))
    return seeq_server_major, seeq_server_minor, seeq_server_patch


def is_spy_module_version_at_least(required_major: int, required_minor: int = 0) -> bool:
    """
    Use this function to ensure that the SPy module meets a version requirement.

    Parameters
    ----------
    required_major : int
        The SPy major version that your notebook/script/application requires.

    required_minor : int, default 0
        The SPy minor version that your notebook/script/application requires.

    Returns
    -------
    True if the SPy version is equal to or greater than the version specified.
    """
    return _is_version_at_least((required_major, required_minor), get_spy_module_version_tuple())


def is_sdk_module_version_at_least(required_major: int, required_minor: int = 0, required_patch: int = 0) -> bool:
    """
    Use this function to ensure that the SDK module meets a version requirement.

    Parameters
    ----------
    required_major : int
        The SDK major version that your notebook/script/application requires.

    required_minor : int, default 0
        The SDK minor version that your notebook/script/application requires.

    required_patch : int, default 0
        The SDK patch version that your notebook/script/application requires.

    Returns
    -------
    True if the SDK version is equal to or greater than the version specified.
    """
    return _is_version_at_least((required_major, required_minor, required_patch), get_sdk_module_version_tuple())


def is_server_version_at_least(required_major: int, required_minor: int = 0, required_patch: int = 0,
                               session: Optional[Session] = None) -> bool:
    """
    Use this function to ensure that the Seeq Server meets a version requirement.

    Parameters
    ----------
    required_major : int
        The Seeq Server major version that your notebook/script/application requires.

    required_minor : int, default 0
        The Seeq Server minor version that your notebook/script/application requires.

    required_patch : int, default 0
        The Seeq Server patch version that your notebook/script/application requires.

    session : spy.Session, optional
        If supplied, the Session object used for the connection to Seeq
        Server. If not supplied, the default session is used.

    Returns
    -------
    True if the Seeq Server version is equal to or greater than the version specified.
    """
    return _is_version_at_least((required_major, required_minor, required_patch), get_server_version_tuple(session))


def _is_version_at_least(required: Tuple, actual: Tuple) -> bool:
    required = list(required)
    actual = list(actual)
    if len(required) > len(actual):
        raise ValueError(f'Cannot compare required version {required} against actual version {actual}. Too many '
                         'version parts specified in required version.')

    while len(required) > 0:
        required_version_part = required.pop(0)
        actual_version_part = actual.pop(0)
        if actual_version_part > required_version_part:
            return True
        elif actual_version_part < required_version_part:
            return False

    return True


SEEQ_SERVER_VERSION_WHERE_SPY_IS_IN_ITS_OWN_PACKAGE = 60


def validate_seeq_server_version(session, status, allow_version_mismatch=False):
    sdk_module_major, sdk_module_minor, _ = get_sdk_module_version_tuple()
    seeq_server_major, seeq_server_minor, seeq_server_patch = get_server_version_tuple(session)

    # The old versioning scheme is like 0.49.3 whereas the new scheme is like 50.1.8
    # See https://seeq.atlassian.net/wiki/spaces/SQ/pages/947225963/Seeq+Versioning+Simplification
    using_old_server_versioning_scheme = (seeq_server_major == 0)

    message = None

    if using_old_server_versioning_scheme:
        if sdk_module_major != seeq_server_major or \
                sdk_module_minor != seeq_server_minor:
            message = (f'The major/minor version of the seeq module ({sdk_module_major}.{sdk_module_minor}) '
                       f'does not match the major/minor version of the Seeq Server you are connected to '
                       f'({seeq_server_major}.{seeq_server_minor}) and is incompatible.')
    else:
        if sdk_module_major != seeq_server_major:
            message = (f'The major version of the seeq module ({sdk_module_major}) '
                       f'does not match the major version of the Seeq Server you are connected to '
                       f'({seeq_server_major}) and is incompatible.')

    if message is not None:
        message += (f'\n\nIt is recommended that you run spy.upgrade() or issue the following PIP command to '
                    f'install a compatible version of the seeq module:\n')
        message += generate_pip_upgrade_command(session, dependencies=[])

        if allow_version_mismatch or session.options.allow_version_mismatch:
            status.warn(message)
        else:
            raise SPyRuntimeError(message)

    return seeq_server_major, seeq_server_minor, seeq_server_patch


def generate_pip_upgrade_command(session: Session, version: Optional[str] = None, use_testpypi: bool = False,
                                 dependencies: Optional[List[str]] = None) -> str:
    sdk_module_major, sdk_module_minor, _ = get_sdk_module_version_tuple()
    seeq_server_major, seeq_server_minor, seeq_server_patch = get_server_version_tuple(session)
    compatible_module_folder = find_compatible_module(session)
    # The old versioning scheme is like 0.49.3 whereas the new scheme is like 50.1.8
    # See https://seeq.atlassian.net/wiki/spaces/SQ/pages/947225963/Seeq+Versioning+Simplification
    using_old_versioning_scheme = (seeq_server_major == 0)
    first_command = None
    if using_old_versioning_scheme:
        install_compatible_sdk = (sdk_module_major != seeq_server_major or sdk_module_minor != seeq_server_minor)
        compatible_sdk_version_specifier = f'{seeq_server_major}.{seeq_server_minor}.{seeq_server_patch}'
    else:
        install_compatible_sdk = (sdk_module_major != seeq_server_major)
        compatible_sdk_version_specifier = f'{seeq_server_major}.{seeq_server_minor}'
    repository_arg = ' --index-url https://test.pypi.org/simple/' if use_testpypi else ''
    dependency_arg = f'{Dependencies(dependencies)}'

    def _install_compatible_sdk():
        if not install_compatible_sdk:
            return None

        if compatible_module_folder is not None:
            return 'pip uninstall -y seeq'
        else:
            return f'pip install -U{repository_arg} seeq~={compatible_sdk_version_specifier}'

    if version is not None:
        if 'r' in version.lower():
            version = re.sub(pattern='r', repl='', string=version, flags=re.IGNORECASE)

        match = re.match(r'^(\d+)\..*', version)
        if not match:
            raise SPyValueError(f'version argument "{version}" is not a full version (e.g. 221.13 or 58.0.2.184.12)')

        version_major = int(match.group(1))
        if version_major < SEEQ_SERVER_VERSION_WHERE_SPY_IS_IN_ITS_OWN_PACKAGE:
            # We're going to the old single-package scheme, where seeq and spy are in the same package and the
            # versioning is something like 58.0.2.184.12. If the currently-installed sdk module is R60 or later, we must
            # uninstall the seeq-spy package so that, when the older seeq package (which includes spy directly) is
            # installed, pip doesn't think that seeq-spy is still installed as well.
            if sdk_module_major >= SEEQ_SERVER_VERSION_WHERE_SPY_IS_IN_ITS_OWN_PACKAGE:
                first_command = 'pip uninstall -y seeq-spy'

            second_command = f'pip install -U{repository_arg} seeq=={version}'
        else:
            # We're going to the new seeq-spy package scheme.
            if seeq_server_major < SEEQ_SERVER_VERSION_WHERE_SPY_IS_IN_ITS_OWN_PACKAGE:
                raise SPyValueError(f'version argument "{version}" is incompatible with Seeq Server version '
                                    f'{session.server_version}')

            first_command = _install_compatible_sdk()
            second_command = f'pip install -U{repository_arg} seeq-spy{dependency_arg}=={version}'
    else:
        if seeq_server_major >= SEEQ_SERVER_VERSION_WHERE_SPY_IS_IN_ITS_OWN_PACKAGE:
            first_command = _install_compatible_sdk()
            second_command = f'pip install -U{repository_arg} seeq-spy{dependency_arg}'
        else:
            if sdk_module_major >= SEEQ_SERVER_VERSION_WHERE_SPY_IS_IN_ITS_OWN_PACKAGE:
                first_command = 'pip uninstall -y seeq-spy'
            second_command = f'pip install -U{repository_arg} seeq~={compatible_sdk_version_specifier}'
    pip_commands = [second_command] if first_command is None else [first_command, second_command]
    pip_command = ' && '.join(pip_commands)
    return pip_command


def find_compatible_module(session: Session):
    """
    Look for a seeq module that is compatible with the version of Seeq Server we're connected to.
    This function is useful in Seeq Data Lab scenarios where the user has installed a "private" version of
    the seeq module (presumably to get a bugfix for SPy) but Seeq Server and Seeq Data Lab have been upgraded in the
    meantime and the user's best course of action is to remove the private version and resume using the "built-in"
    version, which will update with Seeq Data Lab as it is upgraded.
    :return: Path to a compatible module, None if a compatible module is not found.
    """
    seeq_module_major, seeq_module_minor, _ = get_sdk_module_version_tuple()
    seeq_server_major, seeq_server_minor, seeq_server_patch = get_server_version_tuple(session)
    try:
        # Check all installed 'seeq' distributions to find a version-compatible one
        from importlib.metadata import distributions
        for dist in distributions():
            # Use normalized name comparison for safety across Python versions
            if 'Name' in dist.metadata and dist.metadata['Name'].lower() == 'seeq':
                dist_version = dist.version
                if seeq_server_major >= 50 and dist_version.startswith(f'{seeq_server_major}.'):
                    return str(dist.locate_file(''))
                elif seeq_server_major < 50 and dist_version.startswith(f'{seeq_server_major}.{seeq_server_minor}.'):
                    return str(dist.locate_file(''))
    except (ImportError, AttributeError, KeyError):
        pass

    return None



from __future__ import annotations

import json
import os
import re
import textwrap
import warnings
from dataclasses import dataclass
from typing import Dict, Optional, List

import pytz
import requests
from dateutil.tz import tz

from seeq import spy, sdk
from seeq.sdk import *
from seeq.sdk.configuration import ClientConfiguration
from seeq.spy import _url, _common, _datalab, _errors, _version
from seeq.spy._datalab import is_datalab_functions_project
from seeq.spy._errors import *

SEEQ_SDK_RETRY_TIMEOUT_IN_SECONDS_ENV_VAR_NAME = 'SEEQ_SDK_RETRY_TIMEOUT_IN_SECONDS'
DEFAULT_REQUESTS_TIMEOUT = 90


@dataclass(repr=False)
class Session:
    """
    Used to segregate Seeq Server logins and allows for multi-server /
    multi-user concurrent logins. This object encapsulates all server-
    specific state, SPy options and API client configuration.

    Examples
    --------
    Log in to two different servers at the same time:

    >>> session1 = Session()
    >>> session2 = Session()
    >>> spy.login(url='https://server1.seeq.site', username='mark', password='markpassword', session=session1)
    >>> spy.login(url='https://server2.seeq.site', username='alex', password='alexpassword', session=session2)
    """
    datasource_output_cache: dict
    _options: Options = None
    client_configuration: ClientConfiguration = None
    _client: Optional[ApiClient] = None
    _user: Optional[UserOutputV1] = None
    _public_url: Optional[str] = None
    _private_url: Optional[str] = None
    _server_version: Optional[str] = None
    _user_folders: Optional[Dict[str, FolderOutputV1]] = None
    supported_units: Optional[set] = None
    corporate_folder: Optional[FolderOutputV1] = None
    auth_providers: Optional[List[DatasourceOutputV1]] = None
    https_verify_ssl: bool = True
    https_key_file: Optional[str] = None
    https_cert_file: Optional[str] = None

    def __init__(self, options: Options = None, client_configuration: ClientConfiguration = None):
        self.client_configuration = client_configuration if client_configuration is not None else ClientConfiguration()
        self.options = options if options is not None else Options(self.client_configuration)
        self.datasource_output_cache = dict()

        # We have this mechanism so that test_run_notebooks() is able to increase the timeout for the child kernels
        if Session.get_global_sdk_retry_timeout_in_seconds() is not None:
            self.options.retry_timeout_in_seconds = Session.get_global_sdk_retry_timeout_in_seconds()

    def __repr__(self):
        return self.get_info()

    def __hash__(self):
        # Added so that the lru_cache can be used when Session is an argument
        return id(self)

    def __getstate__(self):
        # We can only pickle certain members. This has to mirror __setstate__().
        return self.options

    def __setstate__(self, state):
        self.options = state

    def get_info(self, *, html: bool = False) -> str:
        if self.client:
            url_part = self.public_url
            if self.private_url != self.public_url:
                url_part += f' ({self.private_url})'
            info = f'Logged in to <strong>{url_part}</strong> as <strong>{self.user_string}</strong>.\n'
            info += f'Seeq Server Version: <strong>{spy.server_version}</strong>\n'
        else:
            info = 'Not logged in.\n'

        info += (f'Seeq SDK Module Version: <strong>{sdk.__version__}</strong> @ {os.path.dirname(sdk.__file__)}\n'
                 f'Seeq SPy Module Version: <strong>{spy.__version__}</strong> @ {os.path.dirname(spy.__file__)}')

        if not html:
            info = re.sub(r'<[^>]*?>', '', info)

        return info

    @property
    def user_string(self) -> str:
        if self.user is None:
            return 'Not logged in.'

        user_string = self.user.username
        user_profile = ''
        if self.user.first_name:
            user_profile = self.user.first_name
        if self.user.last_name:
            user_profile += ' ' + self.user.last_name
        if self.user.is_admin:
            user_profile += ' [Admin]'
        if len(user_profile) > 0:
            user_string += ' (%s)' % user_profile.strip()

        return user_string

    @staticmethod
    def validate(session):
        return spy.session if session is None else session

    @staticmethod
    def set_global_sdk_retry_timeout_in_seconds(timeout: Optional[int]):
        """
        This is used to set the SDK's retry timeout (see
        "retry_timeout_in_seconds" in api_client.py) for all
        child Python kernels, such as those spawned by executing
        notebooks via nbformat is in test_run_notebook().
        :param timeout: Timeout (in seconds)

        :meta private:
        """
        if timeout is None and SEEQ_SDK_RETRY_TIMEOUT_IN_SECONDS_ENV_VAR_NAME in os.environ:
            del os.environ[SEEQ_SDK_RETRY_TIMEOUT_IN_SECONDS_ENV_VAR_NAME]
        else:
            os.environ[SEEQ_SDK_RETRY_TIMEOUT_IN_SECONDS_ENV_VAR_NAME] = str(timeout)

    @staticmethod
    def get_global_sdk_retry_timeout_in_seconds() -> Optional[int]:
        """
        See set_global_sdk_retry_timeout_in_seconds()
        :return: Timeout (in seconds)

        :meta private:
        """
        if SEEQ_SDK_RETRY_TIMEOUT_IN_SECONDS_ENV_VAR_NAME in os.environ:
            return int(os.environ[SEEQ_SDK_RETRY_TIMEOUT_IN_SECONDS_ENV_VAR_NAME])
        else:
            return None

    def clear(self):
        """
        Re-initializes the object to a "logged out" state. Note that this
        function does NOT reset API client configuration or SPy options.
        """
        self.client = None
        self.user = None
        self.public_url = None
        self.private_url = None
        self.server_version = None
        self.supported_units = None
        self.corporate_folder = None
        self.auth_providers = None
        self.datasource_output_cache = dict()
        self.https_verify_ssl = True
        self.https_key_file = None
        self.https_cert_file = None
        self._user_folders = None

    # Prior to the advent of Session objects, the spy.client, spy.user and spy.server_version module-level variables
    # were exposed to end-users as a convenience. The setters below copy those (now) Session variables to those
    # legacy module-level locations for backward compatibility purposes. (Only if this Session object is the default
    # Session.)
    @property
    def client(self) -> Optional[ApiClient]:
        """
        Get the API client object for this session
        """
        return self._client

    @client.setter
    def client(self, value):
        self._client = value
        if self._client is not None:
            self._client.user_agent = f'Seeq-Python-SPy/{spy.__version__}/python'
        if self is spy.session:
            spy.client = self._client

    @property
    def user(self) -> Optional[UserOutputV1]:
        """
        Get the user that is logged into this session
        """
        return self._user

    @user.setter
    def user(self, value):
        self._user = value
        if self is spy.session:
            spy.user = self._user

    @property
    def server_version(self) -> Optional[str]:
        """
        Get the version of the Seeq server
        this session is logged into
        """
        return self._server_version

    @server_version.setter
    def server_version(self, value):
        self._server_version = value
        if self is spy.session:
            spy.server_version = self._server_version

    @property
    def public_url(self):
        return self._public_url

    @public_url.setter
    def public_url(self, value):
        self._public_url = _url.cleanse_url(value)

    @property
    def private_url(self):
        return self._private_url

    @private_url.setter
    def private_url(self, value):
        self._private_url = _url.cleanse_url(value)

    def get_api_url(self):
        """
        Returns the URL to use for API calls, which ends up being the
        private URL (if specified) or the public URL.

        :meta private:
        """
        return f'{self.private_url}/api'

    @property
    def options(self):
        """
        Assign a new value to the following variables if you would like to adjust them.

        ``spy.options.compatibility`` (default: None)

            The major version of SPy to emulate from a compatibility standpoint. This
            is important to set if you would like to minimize the chance that your
            script or add-on "breaks" when SPy is upgraded. Set it to the major version
            of SPy that you have tested against. E.g.: spy.options.compatibility = 184

        ``spy.options.search_page_size`` (default: 1000)

            The number of items retrieved on each round-trip to the Seeq Server during
            a spy.search() call. If you have a fast system and fast connection, you can
            make this higher.

        ``spy.options.pull_page_size`` (default: 1000000)

            The number of samples/capsules retrieved on each round-trip to the Seeq
            Server during a spy.pull() call. If you have a slow system or slow
            connection, you may wish to make this lower. It is not recommended to
            exceed 1000000.

        ``spy.options.push_page_size`` (default: 100000)

            The number of samples/capsules uploaded during each round-trip to the Seeq
            Server during a spy.push() call. If you have a slow system or slow
            connection, you may wish to make this lower. It is not recommended to
            exceed 1000000.

        ``spy.options.metadata_push_batch_size`` (default: 1000)

            The number of items uploaded during each round-trip to the Seeq
            Server during a spy.push(metadata) call. If you have a low-memory system
            you may wish to make this lower. It is not recommended to exceed 10000.

        ``spy.options.max_concurrent_requests`` (default: 8)

            The maximum number of simultaneous requests made to the Seeq Server during
            spy.pull() and spy.push() calls. The higher the number, the more you can
            monopolize the Seeq Server. If you keep it low, then other users are less
            likely to be impacted by your activity.

        ``spy.options.retry_timeout_in_seconds`` (default: 5)

            The amount of time to spend retrying a failed Seeq Server API call in an
            attempt to overcome network flakiness.

        ``spy.options.request_timeout_in_seconds`` (default: None)

            The amount of time to wait for a single request to complete, after which
            the http client will consider it is taking too long and give up on it.
            The default of None indicates there is no limit (infinite timeout).

        ``spy.options.clear_content_cache_before_render`` (default: False)

            When using spy.workbooks.pull(include_rendered_content=True), always
            re-render the content even if it had been previously rendered and cached.

        ``spy.options.force_calculated_scalars`` (default: False)

            During spy.push(metadata), always push CalculatedScalars even if
            LiteralScalars would normally apply. (Ignored in R60 and earlier.)

        ``spy.options.allow_version_mismatch`` (default: False)

            Allow a major version mismatch between SPy and Seeq Server. (Normally,
            a mismatch raises a RuntimeError.)

        ``spy.options.friendly_exceptions`` (default: True if running in Data Lab, otherwise False)

            If True, exceptions raised in a Jupyter notebook will be displayed in a
            friendlier format. Stack traces will not be shown by default for most
            errors; error messages will precede the stack trace; and internal SPy
            code will be omitted from the stack trace.

        ``spy.options.default_timezone`` (default: None)

            If set to a timezone, this will be understood as the intended timezone
            for all naive datetimes passed as input to SPy. This will not override
            the timezone of any timezone-aware datetime. If set to None, naive
            datetimes will be interpreted as being in the logged-in user's preferred
            timezone. Timezone can be specified as str, pytz.timezone or dateutil.tzinfo.

        ``spy.options.include_push_method`` (default: False)

            If True, a Push Method column will be added to the metadata push result DataFrame
            that indicates whether an item was pushed via the (fast) ``batch`` REST APIs or had
            to resort to the (slower) ``single`` REST APIs.
        """
        return self._options

    @options.setter
    def options(self, value):
        self._options = value

    @property
    def requests(self):
        return SqAuthRequests(self)

    @property
    def request_origin_label(self) -> str:
        """
        Used for tracking Data Consumption. If supplied, this label will be added as a header to all requests from
        the logged in user. Data Lab will automatically provide a default that you can choose to override.
        """
        return self.client.default_headers.get('x-sq-origin-label')

    @request_origin_label.setter
    def request_origin_label(self, value: str):
        if self.client is None:
            raise RuntimeError("Cannot set request_origin_label before logging in")

        if value is None:
            if 'x-sq-origin-label' in self.client.default_headers:
                del self.client.default_headers['x-sq-origin-label']
            return

        if is_datalab_functions_project() and not value.startswith("[Data Lab Functions]"):
            value = f"[Data Lab Functions] {value}"
        self.client.set_default_header('x-sq-origin-label', value)

    @property
    def request_origin_url(self) -> str:
        """
        Used for tracking Data Consumption. If supplied, this label will be added as a header to all requests from
        the logged in user. Data Lab will automatically provide a default that you can choose to override. If NOT in
        Data Lab, supply a full URL that leads to the tool/plugin that is consuming data, if applicable.
        """
        return self.client.default_headers.get('x-sq-origin-url')

    @request_origin_url.setter
    def request_origin_url(self, value: str):
        if self.client is None:
            raise RuntimeError("Cannot set request_origin_url before logging in")

        if value is None:
            if 'x-sq-origin-url' in self.client.default_headers:
                del self.client.default_headers['x-sq-origin-url']
            return

        self.client.set_default_header('x-sq-origin-url', value)
        self.client.set_default_header('Referer', value)

    def get_user_folder(self, user_id: str) -> FolderOutputV1:
        """
        Get the specified user's home folder. Requires admin permissions.
        """
        if not self.user.is_admin:
            raise SPyRuntimeError('get_user_folders() requires admin permissions')

        if not self.client:
            raise SPyRuntimeError('get_user_folders() requires logged in session')

        if self._user_folders is not None:
            return self._user_folders.get(user_id)

        folders_api = FoldersApi(self.client)
        offset = 0
        limit = self.options.search_page_size
        self._user_folders = dict()
        while True:
            folders_output = folders_api.get_folders(filter='Users', folder_id='users', limit=limit, offset=offset)
            for content in folders_output.content:
                if content.owner is not None:
                    self._user_folders[content.owner.id] = content

            if len(folders_output.content) < limit:
                break

            offset += limit

        return self._user_folders.get(user_id)

    def get_user_timezone(self, default_tz='UTC'):
        """
        Returns the preferred timezone of the user currently logged in, or default_tz if there is no user currently
        logged in.

        :param: session: The login session (necessary to fulfill this call).
        :param: default_tz: The default timezone to return if no user is logged in.
        :return: The user's preferred timezone, in IANA Time Zone Database format (e.g., 'America/New York')
        :rtype: str
        """
        _common.validate_timezone_arg(default_tz)
        try:
            workbench_dict = json.loads(self.user.workbench)
            return workbench_dict['state']['stores']['sqWorkbenchStore']['userTimeZone']
        except (AttributeError, KeyError, TypeError):
            # This can happen if the user has never logged in interactively (e.g., agent_api_key)
            return default_tz

    def get_session_timezone(self):
        """
        Gets the timezone to interpret a datetime as if none is specified. This is the
        spy.options.default_timezone timezone if it exists, else the user's preferred timezone.
        :return: The session timezone, in IANA Time Zone Database format (e.g., 'America/New York')
        :rtype: str
        """
        return self.options.default_timezone if self.options.default_timezone else self.get_user_timezone()

    # The list of units that come back from system_api.get_supported_units() is a curated set of compound units that are
    # meant to be human-friendly in Formula tool help. But it's the only thing available in the API to figure out what
    # base units are supported, so we have to split the compound unit string into its components. I.e., S/cm³ gets split
    # into S and cm so that we can add those two "base" units to the set we use to determine validity.
    UNITS_SPLIT_REGEX = r'[/*²³·]'

    def is_valid_unit(self, unit):
        """
        Returns True if the supplied unit will be recognized by the Seeq calculation engine. This can be an important
        function to use if you are attempting to supply a "Value Unit Of Measure" property on a Signal or a "Unit Of
        Measure" property on a Scalar.

        :param: session: The login session (necessary to execute this call)
        :param: unit: The unit of measure for which to assess validity
        :return: True if unit is valid, False if not
        """
        if not self.supported_units:
            system_api = SystemApi(self.client)
            support_units_output = system_api.get_supported_units()  # type: SupportedUnitsOutputV1

            self.supported_units = set()
            for supported_unit_family in support_units_output.units.values():
                for supported_unit in supported_unit_family:
                    unit_parts = re.split(self.UNITS_SPLIT_REGEX, supported_unit)
                    self.supported_units.update([u for u in unit_parts if len(u) > 0])

        unit_parts = re.split(self.UNITS_SPLIT_REGEX, unit)
        for unit_part in [u for u in unit_parts if len(u) > 0]:
            if unit_part not in self.supported_units:
                return False

        return True


class Options:
    _DEFAULT_SEARCH_PAGE_SIZE = 1_000
    _DEFAULT_PULL_PAGE_SIZE = 1_000_000
    _DEFAULT_TABLE_PULL_PAGE_SIZE = 10_000
    _DEFAULT_PUSH_PAGE_SIZE = 100_000
    _DEFAULT_METADATA_PUSH_BATCH_SIZE = 1_000
    _DEFAULT_MAX_CONCURRENT_REQUESTS = 8
    _DEFAULT_CLEAR_CONTENT_CACHE_BEFORE_RENDER = False
    _DEFAULT_FORCE_CALCULATED_SCALARS = True
    _DEFAULT_ALLOW_VERSION_MISMATCH = False
    _DEFAULT_FRIENDLY_EXCEPTIONS = _datalab.is_datalab()
    _DEFAULT_TIMEZONE = None
    _DEFAULT_COMPATIBILITY = None
    _DEFAULT_MIN_COMPATIBILITY = 188
    _DEFAULT_INCLUDE_PUSH_METHOD = False
    _DEFAULT_STATUS_WARNING_LIMIT = 20

    def __init__(self, client_configuration: ClientConfiguration):
        self.client_configuration = client_configuration
        self.search_page_size = self._DEFAULT_SEARCH_PAGE_SIZE
        self.pull_page_size = self._DEFAULT_PULL_PAGE_SIZE
        self.table_pull_page_size = self._DEFAULT_TABLE_PULL_PAGE_SIZE
        self.push_page_size = self._DEFAULT_PUSH_PAGE_SIZE
        self.metadata_push_batch_size = self._DEFAULT_METADATA_PUSH_BATCH_SIZE
        self.max_concurrent_requests = self._DEFAULT_MAX_CONCURRENT_REQUESTS
        self.clear_content_cache_before_render = self._DEFAULT_CLEAR_CONTENT_CACHE_BEFORE_RENDER
        self.force_calculated_scalars = self._DEFAULT_FORCE_CALCULATED_SCALARS
        self.allow_version_mismatch = self._DEFAULT_ALLOW_VERSION_MISMATCH
        self.default_timezone = self._DEFAULT_TIMEZONE
        self._compatibility = self._DEFAULT_COMPATIBILITY
        self.include_push_method = self._DEFAULT_INCLUDE_PUSH_METHOD
        self.status_warning_limit = self._DEFAULT_STATUS_WARNING_LIMIT
        try:
            self.friendly_exceptions = self._DEFAULT_FRIENDLY_EXCEPTIONS
        except RuntimeError:
            pass

    @property
    def compatibility(self):
        return self._compatibility

    @compatibility.setter
    def compatibility(self, value: Optional[int]):
        _common.validate_argument_types([
            (value, 'compatibility', (int, float)),
        ])
        if value is None:
            self._compatibility = None
            return
        if isinstance(value, float):
            # Users may try to provide a point fix compatibility, but our compatibility flag is on Major only.
            # Drop the decimal values here. Floats don't work like version numbers anyway (190.1 vs 190.10).
            value = int(value)
        max_compatibility, _ = _version.get_spy_module_version_tuple()
        if value < self._DEFAULT_MIN_COMPATIBILITY:
            warnings.warn(f"Compatibility value {value} is below the minimum value {self._DEFAULT_MIN_COMPATIBILITY}. "
                          f"Defaulting to the minimum value.")
            self._compatibility = self._DEFAULT_MIN_COMPATIBILITY
        elif value > max_compatibility:
            warnings.warn(f"Compatibility value {value} is above the maximum value {max_compatibility}. Defaulting to "
                          f"the maximum value.")
            self._compatibility = max_compatibility
        else:
            self._compatibility = value

    def __str__(self):
        return '\n'.join([f"{k}: {v}" for k, v in self.__dict__.items()])

    def __getstate__(self):
        # We can only pickle certain members. This has to mirror __setstate__().
        return (self.compatibility,
                self.search_page_size,
                self.pull_page_size,
                self.push_page_size,
                self.metadata_push_batch_size,
                self.max_concurrent_requests,
                self.clear_content_cache_before_render,
                self.force_calculated_scalars,
                self.allow_version_mismatch,
                self.table_pull_page_size)

    def __setstate__(self, state):
        def get(i, default=None):
            return state[i] if i < len(state) else default

        # Accessing by index with a default allows us to maintain compatibility with previously-pickled Status
        # DataFrames
        self.compatibility = get(0, self._DEFAULT_COMPATIBILITY)
        self.search_page_size = get(1, self._DEFAULT_SEARCH_PAGE_SIZE)
        self.pull_page_size = get(2, self._DEFAULT_PULL_PAGE_SIZE)
        self.push_page_size = get(3, self._DEFAULT_PUSH_PAGE_SIZE)
        self.metadata_push_batch_size = get(4, self._DEFAULT_METADATA_PUSH_BATCH_SIZE)
        self.max_concurrent_requests = get(5, self._DEFAULT_MAX_CONCURRENT_REQUESTS)
        self.clear_content_cache_before_render = get(6, self._DEFAULT_CLEAR_CONTENT_CACHE_BEFORE_RENDER)
        self.force_calculated_scalars = get(7, self._DEFAULT_FORCE_CALCULATED_SCALARS)
        self.allow_version_mismatch = get(8, self._DEFAULT_ALLOW_VERSION_MISMATCH)
        self.table_pull_page_size = get(9, self._DEFAULT_TABLE_PULL_PAGE_SIZE)

    @property
    def friendly_exceptions(self):
        return self._friendly_exceptions

    @friendly_exceptions.setter
    def friendly_exceptions(self, value):
        if value:
            try:
                _errors.add_spy_exception_handler()
                self._friendly_exceptions = True
            except RuntimeError:
                self._friendly_exceptions = False
                raise

        else:
            _errors.remove_spy_exception_handler()
            self._friendly_exceptions = False

    @property
    def default_timezone(self):
        return self._default_timezone

    @default_timezone.setter
    def default_timezone(self, value):
        if value is None:
            self._default_timezone = None
        elif isinstance(value, str):
            try:
                pytz.timezone(value)
                self._default_timezone = value
            except pytz.UnknownTimeZoneError:
                raise
        elif isinstance(value, pytz.BaseTzInfo) or isinstance(value, tz.tzoffset):
            self._default_timezone = value
        else:
            raise SPyTypeError(f"Default timezone can't be type {type(value).__name__}")

    @property
    def retry_timeout_in_seconds(self):
        return self.client_configuration.retry_timeout_in_seconds

    @retry_timeout_in_seconds.setter
    def retry_timeout_in_seconds(self, value):
        if not isinstance(value, (int, float)) or value <= 0:
            raise SPyValueError("retry_timeout_in_seconds must be a positive, non-zero integer or float number.")

        if self.request_timeout_in_seconds is not None and value > self.request_timeout_in_seconds:
            raise SPyValueError("retry_timeout_in_seconds cannot be more than request_timeout_in_seconds. Set "
                                "request_timeout_in_seconds to an equal or greater value first (or None so that it is "
                                "infinite.")

        self.client_configuration.retry_timeout_in_seconds = value

    @property
    def request_timeout_in_seconds(self):
        # We need to check here because SPy can be used with older versions of Seeq, which do not have this attribute
        if hasattr(self.client_configuration, 'request_timeout_in_seconds'):
            return self.client_configuration.request_timeout_in_seconds
        return None

    @request_timeout_in_seconds.setter
    def request_timeout_in_seconds(self, value):
        if value is not None and (not isinstance(value, (int, float)) or value <= 0):
            raise SPyValueError("request_timeout_in_seconds must be a positive, non-zero integer or float number "
                                "OR None to indicate infinite.")

        if value is not None and value < self.client_configuration.retry_timeout_in_seconds:
            raise SPyValueError("request_timeout_in_seconds cannot be less than retry_timeout_in_seconds. Set "
                                "retry_timeout_in_seconds to an equal or lower value first.")

        # We need to check here because SPy can be used with older versions of Seeq, which do not have this attribute
        if hasattr(self.client_configuration, 'request_timeout_in_seconds'):
            self.client_configuration.request_timeout_in_seconds = value
        else:
            warnings.warn("The request_timeout_in_seconds option is not available in the current version of the "
                          "seeq Python package. Please upgrade to the latest version of the seeq package that "
                          "corresponds to the version of your Seeq service.")

    def print(self):
        _common.print_output(str(self))

    def help(self):
        help_string = f"""\
            Assign a new value to the following variables if you would like to adjust them.

            E.g.:
               spy.options.max_concurrent_requests = 3

            Available Options
            -----------------

            spy.options.compatibility (default: {self._DEFAULT_COMPATIBILITY})

                The major version of SPy to emulate from a compatibility standpoint. This
                is important to set if you would like to minimize the chance that your
                script or add-on "breaks" when SPy is upgraded. Set it to the major version
                of SPy that you have tested against. E.g.: spy.options.compatibility = 184

            spy.options.search_page_size (default: {self._DEFAULT_SEARCH_PAGE_SIZE})

                The number of items retrieved on each round-trip to the Seeq Server during
                a spy.search() call. If you have a fast system and fast connection, you can
                make this higher.

            spy.options.pull_page_size (default: {self._DEFAULT_PULL_PAGE_SIZE})

                The number of samples/capsules retrieved on each round-trip to the Seeq
                Server during a spy.pull() call. If you have a slow system or slow
                connection, you may wish to make this lower. It is not recommended to
                exceed 1000000.

            spy.options.table_pull_page_size (default: {self._DEFAULT_TABLE_PULL_PAGE_SIZE})

                The number of rows retrieved on each round-trip to the Seeq Server
                during a spy.pull() call where a Vantage Room is being pulled. If
                you have a slow system or slow connection, you may wish to make this
                lower. It is not recommended to exceed 10_000.

            spy.options.push_page_size (default: {self._DEFAULT_PUSH_PAGE_SIZE})

                The number of samples/capsules uploaded during each round-trip to the Seeq
                Server during a spy.push() call. If you have a slow system or slow
                connection, you may wish to make this lower. It is not recommended to
                exceed 1000000.

            spy.options.metadata_push_batch_size (default: {self._DEFAULT_METADATA_PUSH_BATCH_SIZE})

                The number of items uploaded during each round-trip to the Seeq
                Server during a spy.push(metadata) call. If you have a low-memory system
                you may wish to make this lower. It is not recommended to exceed 10000.

            spy.options.max_concurrent_requests (default: {self._DEFAULT_MAX_CONCURRENT_REQUESTS})

                The maximum number of simultaneous requests made to the Seeq Server during
                spy.pull() and spy.push() calls. The higher the number, the more you can
                monopolize the Seeq Server. If you keep it low, then other users are less
                likely to be impacted by your activity.

            spy.options.retry_timeout_in_seconds (default: {ClientConfiguration.DEFAULT_RETRY_TIMEOUT_IN_SECONDS})

                The amount of time to spend retrying a failed Seeq Server API call in an
                attempt to overcome network flakiness.

            spy.options.request_timeout_in_seconds (default: None)

                The amount of time to wait for a single request to complete, after which
                the http client will consider it is taking too long and give up on it.
                The default of None indicates there is no limit (infinite timeout).

                This option may require an upgrade of the `seeq` Python package to
                function - a warning will be provided if so.

            spy.options.clear_content_cache_before_render (default: {str(self._DEFAULT_CLEAR_CONTENT_CACHE_BEFORE_RENDER)})

                When using spy.workbooks.pull(include_rendered_content=True), always
                re-render the content even if it had been previously rendered and cached.

            spy.options.force_calculated_scalars (default: {self._DEFAULT_FORCE_CALCULATED_SCALARS})

                During spy.push(metadata), always push CalculatedScalars even if
                LiteralScalars would normally apply. (Ignored in R60 and earlier.)

            spy.options.allow_version_mismatch (default: {self._DEFAULT_ALLOW_VERSION_MISMATCH})

                Allow a major version mismatch between SPy and Seeq Server. (Normally,
                a mismatch raises a RuntimeError.)

            spy.options.friendly_exceptions (default: True if running in Data Lab, otherwise False)

                If True, exceptions raised in a Jupyter notebook will be displayed in a
                friendlier format. Stack traces will not be shown by default for most
                errors; error messages will precede the stack trace; and internal SPy
                code will be omitted from the stack trace.

            spy.options.default_timezone (default: {self._DEFAULT_TIMEZONE})

                If set to a timezone, this will be understood as the intended timezone
                for all naive datetimes passed as input to SPy. This will not override
                the timezone of any timezone-aware datetime. If set to None, naive
                datetimes will be interpreted as being in the logged-in user's preferred
                timezone. Timezone can be specified as str, pytz.timezone or dateutil.tzinfo.

            spy.options.include_push_method (default: {self._DEFAULT_INCLUDE_PUSH_METHOD})

                If True, a Push Method column will be added to the metadata push result DataFrame
                that indicates whether an item was pushed via the (fast) ``batch`` REST APIs or had
                to resort to the (slower) ``single`` REST APIs.

            spy.options.status_warning_limit (default: {self._DEFAULT_STATUS_WARNING_LIMIT})
                The number of warnings to display in the status output before truncating.
                If the number of warnings exceeds this limit, a message will be displayed
                indicating how many warnings were truncated.
        """

        _common.print_output(textwrap.dedent(help_string))

    def wants_compatibility_with(self, major_version):
        return self.compatibility is not None and self.compatibility <= major_version


@dataclass(repr=False)
class SqAuthRequests:
    session: Session

    def add_request_kwargs(self, d):
        if 'cookies' not in d:
            d['cookies'] = dict()
        d['cookies']['sq-auth'] = self.session.client.auth_token
        if 'headers' not in d:
            d['headers'] = dict()
        d['headers']['x-sq-auth'] = self.session.client.auth_token
        d['verify'] = self.session.https_verify_ssl
        if 'timeout' not in d:
            d['timeout'] = DEFAULT_REQUESTS_TIMEOUT

    def get(self, *args, **kwargs) -> requests.Response:
        self.add_request_kwargs(kwargs)
        return requests.get(*args, **kwargs)

    def patch(self, *args, **kwargs) -> requests.Response:
        self.add_request_kwargs(kwargs)
        return requests.patch(*args, **kwargs)

    def post(self, *args, **kwargs) -> requests.Response:
        self.add_request_kwargs(kwargs)
        return requests.post(*args, **kwargs)

    def put(self, *args, **kwargs) -> requests.Response:
        self.add_request_kwargs(kwargs)
        return requests.put(*args, **kwargs)

    def delete(self, *args, **kwargs) -> requests.Response:
        self.add_request_kwargs(kwargs)
        return requests.delete(*args, **kwargs)



import re
from seeq import spy
from seeq.sdk import *


def get_default_page_size():
    try:
        system_api = SystemApi(spy.client)
        system_settings = system_api.get_server_status(
            include_internal=True
        )  # type: ServerStatusOutputV1
        specs = system_settings.server_specs

        for spec in specs:
            if spec.component_name == "RecommendedPageSize":
                page_size = re.sub("[^0-9]", "", spec.system_spec_value)
                if page_size.isnumeric():
                    return int(page_size)
    except Exception as e:
        print("Seeq Server error encountered: %s" % e)

    # return a default value if we
    # could not get the page size
    return 10000



from seeq.sdk import ApiClient
from seeq.sdk.api import UsersApi, UserGroupsApi, FormulasApi
from seeq.sdk.models import PutScalarsInputV1, UserOutputV1, UserGroupOutputV1, AddOnToolInputV1, WorkbookOutputV1
from seeq.spy import _version


def is_continuation_token_used() -> bool:
    """
    Check if compute API endpoints should use the continuation_token field for pagination. On false, should use the
    old start key increment and deduplicate capsules scheme.

    Returns
    -------
    True if the SDK version is equal to or greater than 63.
    """
    return _version.is_sdk_module_version_at_least(63)


def is_force_calculated_scalars_available() -> bool:
    """
    Check if the force_calculated_scalars parameter is available on the PutScalars endpoint.

    Returns
    -------
    True if the force_calculated_scalars parameter is available on the PutScalars endpoint.
    """
    return 'force_calculated_scalars' in PutScalarsInputV1.attribute_map


def is_launch_location_available() -> bool:
    """
    Check if the launch_location parameter is available on the AddOnTools endpoint.

    Returns
    -------
    True if the launch_location parameter is available on the AddOnTools endpoint.
    """
    return 'launch_location' in AddOnToolInputV1.attribute_map


def is_compile_formula_and_parameters_available() -> bool:
    """
    Check if the compile_formula_and_parameters method is available in the FormulasApi.

    Returns
    -------
    True if the compile_formula_and_parameters method is available in the FormulasApi.
    """
    return hasattr(FormulasApi, 'compile_formula_and_parameters')


def is_vantage_supported() -> bool:
    """
    Check if Vantage is available in this server.

    Returns
    -------
    True if Vantage is supported.
    """
    return _version.is_sdk_module_version_at_least(66, 30, 0)


def are_vantage_rooms_as_workbooks_supported() -> bool:
    """
    Vantage Rooms used to be in their own weird database. Now they look like workbooks in the backend and have
    an evidence_table_definition_id field.

    Returns
    -------
    True if Vantage Rooms are backed by workbooks.
    """
    return 'evidence_table_definition_id' in WorkbookOutputV1.attribute_map


def get_user(client: ApiClient, user_id: str, *, include_groups: bool = True) -> UserOutputV1:
    """
    Get a user

    Parameters
    ----------
    client : ApiClient
        The api_client to use for the SDK call.
    user_id : str
        ID of the user to get (required)
    include_groups : bool
        Include the groups of which the user is a member. Note that, depending on the version of Seeq, this parameter
        may not be supported.
    """
    users_api = UsersApi(client)
    try:
        return users_api.get_user(id=user_id, include_groups=include_groups)
    except TypeError:
        return users_api.get_user(id=user_id)


def get_user_group(client: ApiClient, user_group_id: str, *, include_members: bool = True) -> UserGroupOutputV1:
    """
    Get a user group.

    client : ApiClient
        The session to use for the request (required)
    user_group_id : str
        ID of the user group to get (required)
    include_members : bool
        Include the members (users) of the user group. Note that, depending on the version of Seeq, this parameter may
        not be supported.
    """
    usergroups_api = UserGroupsApi(client)
    try:
        return usergroups_api.get_user_group(user_group_id=user_group_id, include_members=include_members)
    except TypeError:
        return usergroups_api.get_user_group(user_group_id=user_group_id)



from __future__ import annotations

import inspect
import sys
import types
from collections.abc import Mapping, Sequence, Set as ABCSet
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Literal,
    Union,
    get_args,
    get_origin,
    get_type_hints,
    TypeVar,
)

from seeq.spy._errors import SPyTypeError

_NONE_TYPE = type(None)


def _safe_get_type_hints(func: Callable) -> Dict[str, object]:
    # Get the module where the function is defined to use its globals for evaluation
    # This "globalns" variable is needed for the get_type_hints() function
    # noinspection PyUnresolvedReferences
    globalns = getattr(sys.modules.get(func.__module__, None), '__dict__', {})

    try:
        return get_type_hints(func, globalns=globalns, include_extras=True)  # py>=3.9
    except TypeError as e:
        # Python 3.8 doesn't have include_extras parameter
        try:
            return get_type_hints(func, globalns=globalns)  # py<=3.8 behavior
        except Exception as inner_e:
            # noinspection PyUnresolvedReferences
            raise SPyTypeError(
                f"Failed to resolve type hints for function {func.__module__}.{func.__qualname__}. "
                f"This is likely due to missing imports or forward references that can't be evaluated. "
                f"Original error: {inner_e}"
            ) from inner_e
    except Exception as e:
        # noinspection PyUnresolvedReferences
        raise SPyTypeError(
            f"Failed to resolve type hints for function {func.__module__}.{func.__qualname__}. "
            f"This is likely due to missing imports or forward references that can't be evaluated. "
            f"Original error: {e}"
        ) from e


def _is_runtime_type(tp: object) -> bool:
    try:
        return isinstance(tp, type)
    except Exception:
        return False


def _type_name(tp: object, for_union: bool = False) -> str:
    try:
        # types and classes
        if _is_runtime_type(tp):
            return tp.__name__  # type: ignore[attr-defined]

        # Special handling for Union types
        origin = get_origin(tp)
        args = get_args(tp)
        if origin is Union and args:
            # Filter out NoneType for readability
            non_none_args = [arg for arg in args if arg not in (_NONE_TYPE, type(None))]

            # If only one non-None type remains, just show that type
            if len(args) == 2 and len(non_none_args) == 1:
                # This is Optional[T], return just T's name
                return _type_name(non_none_args[0], for_union=for_union)

            # For Union with multiple non-None types, format as "X or Y or Z"
            if len(non_none_args) > 1:
                type_names = []
                for arg in non_none_args:
                    # When formatting union alternatives, always show full generic types
                    name = _type_name(arg, for_union=True)
                    type_names.append(name)
                return ' or '.join(type_names)

        # For generic types like List[str], Dict[str, int]
        if origin is not None and _is_runtime_type(origin):
            if for_union and args:
                # Show full generic type for Union error messages
                result = str(tp).replace('typing.', '')
                return result
            else:
                # Show just the origin type name for container mismatch errors
                return origin.__name__  # type: ignore[attr-defined]

        # Convert to string and clean up typing module prefixes for readability
        result = str(tp)
        # For Optional[X], strip Optional wrapper to show just X
        if 'Optional[' in result:
            # Extract the inner type from Optional[X]
            start = result.find('Optional[') + 9
            # Find matching closing bracket
            depth = 1
            i = start
            while i < len(result) and depth > 0:
                if result[i] == '[':
                    depth += 1
                elif result[i] == ']':
                    depth -= 1
                i += 1
            if depth == 0:
                result = result[start:i - 1]
        result = result.replace('typing.', '')
        return result
    except Exception:
        return repr(tp)


def _format_value_type(value: object) -> str:
    try:
        return type(value).__name__
    except Exception:
        return "<unknown>"


def _append(errors: List[str], path: str, expected: object, value: object, why: Union[str, None] = None,
            for_union: bool = False) -> None:
    msg = f"Argument '{path}' should be type {_type_name(expected, for_union=for_union)}, but is type {_format_value_type(value)}"
    if why:
        msg += f" ({why})"
    errors.append(msg)


def _check(value: object, expected: object, path: str, errors: List[str]) -> None:
    # No annotation / Any => accept
    if expected is inspect._empty or expected is Any:
        return

    # Handle typing.ForwardRef already resolved by get_type_hints in normal cases.
    # Still, guard against odd cases.
    if expected is None:
        if value is not None:
            _append(errors, path, expected, value)
        return

    # NewType support: expected.__supertype__
    supertype = getattr(expected, "__supertype__", None)
    if supertype is not None:
        _check(value, supertype, path, errors)
        return

    origin = get_origin(expected)
    args = get_args(expected)

    # Annotated[T, ...] => validate T
    annotated_type = getattr(__import__("typing"), "Annotated", None)
    if annotated_type is not None and origin is annotated_type:
        if args:
            _check(value, args[0], path, errors)
        return

    # Literal
    if origin is Literal:
        if value not in args:
            _append(errors, path, expected, value, why=f"allowed values: {args!r}")
        return

    # TypeVar
    if isinstance(expected, TypeVar):
        # If constrained, accept if matches any constraint
        if expected.__constraints__:
            for c in expected.__constraints__:
                tmp: List[str] = []
                _check(value, c, path, tmp)
                if not tmp:
                    return
            _append(errors, path, expected, value, why=f"constraints: {expected.__constraints__!r}")
            return
        # If bounded, validate against the bound
        if expected.__bound__ is not None:
            _check(value, expected.__bound__, path, errors)
        # Else unconstrained TypeVar => can't validate
        return

    # Union / Optional / PEP 604 (A | B)
    # In 3.10+, get_origin(int | str) is types.UnionType; in typing.Union it's typing.Union
    union_types = tuple(
        t for t in (getattr(types, "UnionType", None), getattr(__import__("typing"), "Union", None)) if t is not None)
    if origin in union_types:
        # Accept if any branch accepts
        best_errors: List[str] = []
        for alt in args:
            tmp: List[str] = []
            _check(value, alt, path, tmp)
            if not tmp:
                return
            # Keep track of the best (most specific) error - prefer errors from non-None types
            if not best_errors or (len(tmp) <= len(best_errors) and alt is not _NONE_TYPE):
                best_errors = tmp

        # If the best error is more specific (has a path like "items[0]"), use it
        # Otherwise, generate a Union-level error message
        if best_errors and '[' in best_errors[0]:
            # This is a nested error (e.g., "items[0] should be..."), use it as-is
            errors.extend(best_errors)
        else:
            # Top-level Union mismatch, show all alternatives in error message
            _append(errors, path, expected, value, for_union=True)
        return

    # NoneType
    if expected is _NONE_TYPE:
        if value is not None:
            _append(errors, path, expected, value)
        return

    # typing.Type[T] => a class object that is subclass of T
    if origin is type:
        if not isinstance(value, type):
            _append(errors, path, expected, value, why="value is not a class/type")
            return
        # If parametrized, enforce subclass check
        if args:
            target = args[0]
            target_origin = get_origin(target) or target
            if _is_runtime_type(target_origin):
                if not issubclass(value, target_origin):  # type: ignore[arg-type]
                    _append(errors, path, expected, value, why=f"not a subclass of {_type_name(target_origin)}")
            # else: can't strict-check
        return

    # Callable[...] => only check callability (signature/arg types are not reliably enforceable at runtime)
    if origin is Callable or expected is Callable:
        if not callable(value):
            _append(errors, path, expected, value, why="not callable")
        return

    # Builtin generics: list[T], set[T], frozenset[T]
    if origin in (list, set, frozenset):
        if not isinstance(value, origin):
            _append(errors, path, expected, value)
            return
        elem_t = args[0] if args else Any
        for i, v in enumerate(value):  # type: ignore[assignment]
            _check(v, elem_t, f"{path}[{i}]", errors)
        return

    # tuple[T1, T2] or tuple[T, ...]
    if origin is tuple:
        if not isinstance(value, tuple):
            _append(errors, path, expected, value)
            return
        if not args:
            return
        # tuple[T, ...]
        if len(args) == 2 and args[1] is Ellipsis:
            elem_t = args[0]
            for i, v in enumerate(value):
                _check(v, elem_t, f"{path}[{i}]", errors)
            return
        # tuple[T1, T2, ...]
        if len(value) != len(args):
            _append(errors, path, expected, value, why=f"tuple length {len(value)} != {len(args)}")
            return
        for i, (v, t) in enumerate(zip(value, args)):
            _check(v, t, f"{path}[{i}]", errors)
        return

    # dict[K, V]
    if origin is dict:
        if not isinstance(value, dict):
            _append(errors, path, expected, value)
            return
        k_t, v_t = args if len(args) == 2 else (Any, Any)
        for k, v in value.items():
            _check(k, k_t, f"{path}.keys()", errors)
            _check(v, v_t, f"{path}[{k!r}]", errors)
        return

    # collections.abc generics (Mapping[K,V], Sequence[T], Set[T], etc.)
    # We do a best-effort isinstance check against the origin, then optionally validate elements.
    if origin is not None:
        try:
            if not isinstance(value, origin):  # type: ignore[arg-type]
                _append(errors, path, expected, value)
                return
        except TypeError:
            # Not runtime-checkable (e.g., Protocol without @runtime_checkable)
            # Best-effort: don't fail hard.
            return

        # Element checking for common ABCs:
        # Mapping[K,V]
        if isinstance(value, Mapping):
            if len(args) >= 2:
                k_t, v_t = args[0], args[1]
                for k, v in value.items():
                    _check(k, k_t, f"{path}.keys()", errors)
                    _check(v, v_t, f"{path}[{k!r}]", errors)
            return

        # Avoid treating str/bytes as sequences of elements unless explicitly annotated as such
        if isinstance(value, (str, bytes, bytearray)):
            return

        # Sequence[T]
        if isinstance(value, Sequence):
            if args:
                elem_t = args[0]
                # Only iterate if it's finite / indexable; Sequence is indexable by contract.
                for i, v in enumerate(value):
                    _check(v, elem_t, f"{path}[{i}]", errors)
            return

        # Set[T]
        if isinstance(value, ABCSet):
            if args:
                elem_t = args[0]
                for v in value:
                    _check(v, elem_t, f"{path}[*]", errors)
            return

    # Plain class/type
    if _is_runtime_type(expected):
        # Special case: bool is subclass of int in Python, but for type validation
        # we want to treat them as distinct types
        if expected is int and isinstance(value, bool):
            _append(errors, path, expected, value)
            return
        if not isinstance(value, expected):  # type: ignore[arg-type]
            _append(errors, path, expected, value)
        return

    # Last resort: try isinstance, but never crash due to typing objects
    try:
        if not isinstance(value, expected):  # type: ignore[arg-type]
            _append(errors, path, expected, value)
    except TypeError:
        # Not runtime-checkable; skip
        return


def get_spy_function_name(func: Callable) -> str:
    # noinspection PyUnresolvedReferences
    function_name = f"{func.__module__}.{func.__qualname__}"
    function_name_parts = function_name.split(".")
    function_name_parts_filtered = [n for n in function_name_parts if n != "seeq" and not n.startswith('_')]
    return ".".join(function_name_parts_filtered)


def validate_argument_types(func: Callable, args: tuple, kwargs: Dict) -> None:
    if kwargs is None:
        kwargs = {}

    sig = inspect.signature(func)
    hints = _safe_get_type_hints(func)

    # Bind args/kwargs to the signature (this enforces correct call shape)
    bound = sig.bind(*args, **kwargs)
    bound.apply_defaults()

    # Check for missing type hints (developer error detection)
    missing_hints = []
    for name, param in sig.parameters.items():
        # Skip 'self' and 'cls' parameters (they don't need type hints)
        if name in ('self', 'cls'):
            continue

        # Get the annotation for this parameter
        annotation = hints.get(name,
                               param.annotation if param.annotation is not inspect.Parameter.empty else inspect._empty)

        # Check if the parameter is missing a type hint
        if annotation is inspect._empty:
            missing_hints.append(name)

    if missing_hints:
        # noinspection PyUnresolvedReferences
        raise TypeError(
            f"Function {func.__module__}.{func.__qualname__} is missing type hints for parameters: {', '.join(missing_hints)}. "
            f"All parameters must have type hints for validation to work properly."
        )

    errors: List[str] = []

    log_message = [f'{get_spy_function_name(func)}() called with arguments:']

    # Validate each bound argument
    for name, value in bound.arguments.items():
        try:
            log_message.append(f'    {name} = {value!r}')
        except Exception:
            log_message.append(f'    {name} = <unrepresentable value> (type: {_format_value_type(value)})')

        expected = hints.get(name, sig.parameters[name].annotation if name in sig.parameters else inspect._empty)

        # If parameter is *args / **kwargs, validate appropriately:
        param = sig.parameters.get(name)
        if param is not None and param.kind is inspect.Parameter.VAR_POSITIONAL:
            # Annotation is the element type for *args
            elem_expected = expected
            for i, v in enumerate(value):
                _check(v, elem_expected, f"*{name}[{i}]", errors)
            continue

        if param is not None and param.kind is inspect.Parameter.VAR_KEYWORD:
            # Annotation is the value type for **kwargs
            val_expected = expected
            for k, v in value.items():
                _check(v, val_expected, f"**{name}[{k!r}]", errors)
            continue

        _check(value, expected, name, errors)

    if errors:
        # Raise the first error as SPyTypeError (for backward compatibility with test expectations)
        raise SPyTypeError(errors[0])

    if bound.arguments.get('status') is not None:
        bound.arguments['status'].log("\n".join(log_message))
